from __future__ import absolute_import
from .asset import Asset


class FlightPhase(Asset):
	"""
	Retrieve the data of defined flight phases.
	"""

	def __init__(self, conn, ems_id=7):
		"""
		Flight phase object initialization

		Parameters
		----------
		conn: emspy.connection.Connection
			connection object
		ems_id: int
			EMS system id
		"""
		Asset.__init__(self, conn, "FlightPhase")
		self._ems_id = ems_id
		self.update_list()

	def update_list(self):
		"""
		Update flight phase list

		Returns
		-------
		None
		"""
		Asset.update_list(self, uri_keys=('flt_phase','list'), uri_args=self._ems_id)
		self._rename_datacol('description', 'name')

	def get_id(self, name=None):
		"""
		Get flight phase id from name

		Parameters
		----------
		name: str
			flight phase name

		Returns
		-------
		int
			flight phase id
		"""
		flightphase_id = self.search('name', name)['id'].tolist()
		return flightphase_id if len(flightphase_id) > 1 else flightphase_id[0]

	def get_name(self, id_val=None):
		"""
		Get flight phase name from id

		Parameters
		----------
		id_val: int
			flight phase id

		Returns
		-------
		str
			flight phase name
		"""
		flightphase_name = self.search('id', id_val, searchtype="match")['name'].tolist()
		return flightphase_name if len(flightphase_name) > 1 else flightphase_name[0]



from __future__ import absolute_import
from .asset import Asset


class Aircraft(Asset):
    """ Manages aircraft info """

    def __init__(self, conn, ems_id):
        """
        Aircraft asset object initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        ems_id: int
            EMS system id
        """
        Asset.__init__(self, conn, "Aircraft")
        self._ems_id = ems_id
        self.update_list()

    def update_list(self):
        """
        Update aircraft list

        Returns
        -------
        None
        """
        Asset.update_list(self, uri_keys=('aircraft', 'list'), uri_args=self._ems_id)
        self._rename_datacol('description', 'name')

    def get_id(self, name=None):
        """
        Get aircraft id from name

        Parameters
        ----------
        name: str
            aircraft name

        Returns
        -------
        aircraft_id: int
            aircraft id
        """
        aircraft_id = self.search('name', name)['id'].tolist()
        return aircraft_id if len(aircraft_id) > 1 else aircraft_id[0]

    def get_name(self, id_val=None):
        """
        Get aircraft name from id

        Parameters
        ----------
        id_val: str
            aircraft id

        Returns
        -------
        aircraft_name: str
            aircraft name
        """
        aircraft_name = self.search('id', id_val, searchtype="match")['name'].tolist()
        return aircraft_name if len(aircraft_name) > 1 else aircraft_name[0]

    def search_by_fleetid(self, fleetid):
        """
        Search aircraft by fleet id

        Parameters
        ----------
        fleetid: str
            fleet id

        Returns
        -------
        pd.DataFrame
            aircrafts that match fleet id
        """
        assets = self.list_all()
        idx = [row.id for row in assets.itertuples() if fleetid in row.fleetIds]
        return assets[assets['id'].isin(idx)]



from builtins import object
from emspy.query import *


class Query(object):
	def __init__(self, conn, ems_name):
		"""
		Query initialization

		Parameters
		----------
		conn: empsy.connection.Connection
			connection object
		ems_name: str
			EMS system name
		"""
		self._conn = conn
		self._ems_name = ems_name
		self.__ems = None
		self.__fleet = None
		self.__aircraft = None
		self.__airport = None
		self.__flightphase = None
		self.__init_ems()
		self._ems_id = self.__ems.get_id(ems_name)
		# self._init_assets()
		# self.reset()

	def __init_ems(self):
		self.__ems = EMS(self._conn)

	def get_ems_id(self):
		"""
		Returns EMS id

		Returns
		-------
		int
			EMS system id
		"""
		return self._ems_id

	def _init_assets(self):
		# self.__fleet = Fleet(self._conn, self._ems_id)
		# self.__aircraft = Aircraft(self._conn, self._ems_id)
		# self.__airport = Airport(self._conn, self._ems_id)
		# self.__flightphase = FlightPhase(self._conn, self._ems_id)		
		# self.__flight = Flight(self._conn, self._ems_id)
		raise NotImplementedError()



from __future__ import print_function
from builtins import object
from emspy.query import LocalData

import networkx as nx
import pandas as pd
import os, sys, re


class Flight(object):
    temp_exclude = [
        'Download Information',
        'Download Review',
        'Processing',
        'Profile 16 Extra Data',
        'Flight Review',
        'Data Information',
        'Operational Information',
        'Operational Information (ODW2)',
        'Weather Information',
        'Profiles',
        'Profile'
    ]

    def __init__(self, conn, ems_id, data_file=LocalData.default_data_file, searchtype='contain'):
        """
        Flight object initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        ems_id: int
            EMS system id
        data_file: str
            path to database file
        searchtype: str
            search type to perform:
                contain: uses the Pandas string method 'contains'
                match: uses the Pandas string method 'match'
        """
        self.searchtype = searchtype
        self._conn = conn
        self._ems_id = ems_id
        self._db_id = None
        self._metadata = None
        self._trees = {'fieldtree': None, 'dbtree': None, 'kvmaps': None}
        self._fields = []
        self.__cntr = 0
        self._uri_root = conn._uri_root

        # Retrieve the field tree data from local storage. If it doesn't exist, generate a new
        # default one.
        self.load_tree(data_file)
        # Set default database
        # self.set_database('fdw flight')

    def load_tree(self, file_name=LocalData.default_data_file):
        """
        Load trees into the object

        Parameters
        ----------
        file_name: str
            path to database file

        Returns
        -------
        None
        """
        if self._metadata is None:
            self._metadata = LocalData(file_name)
        elif not self._metadata.is_db_path_correct(file_name):
            self._metadata.close()
            self._metadata = LocalData(file_name)

        self._trees = {
            'fieldtree': self.__get_fieldtree(),
            'dbtree': self.__get_dbtree(),
            'kvmaps': self.__get_kvmaps()
        }

    def save_tree(self, file_name=LocalData.default_data_file):
        """
        Method to save trees to a defined database file

        Parameters
        ----------
        file_name: str
            path to database file

        Returns
        -------
        None
        """
        from shutil import copyfile
        ld = self._metadata
        if not ld.is_db_path_correct(file_name):
            # A different location is given, if a previous file exists copy it to the new location.
            if ld.file_loc() is not None and file_name is not None:
                copyfile(ld.file_loc(), os.path.abspath(file_name))
            self._metadata.close()
            self._metadata = LocalData(file_name)
        self.__save_fieldtree()
        self.__save_dbtree()
        self.__save_kvmaps()

    def __get_fieldtree(self):
        if self._db_id is None:
            return pd.DataFrame(columns=self._metadata.table_info['fieldtree'])
        else:
            return self._metadata.get_data(
                "fieldtree",
                "ems_id = %d and db_id = '%s'" % (self._ems_id, self._db_id)
            )

    def __save_fieldtree(self):
        if len(self._trees['fieldtree']) > 0:
            self._metadata.delete_data(
                "fieldtree",
                "ems_id = %d and db_id = '%s'" % (self._ems_id, self._db_id)
            )
            self._metadata.append_data("fieldtree", self._trees['fieldtree'])

    def __get_dbtree(self):
        tree = self._metadata.get_data("dbtree", self.__get_filter('dbtree'))
        if len(tree) < 1:
            dbroot = {
                'uri_root': self._uri_root,
                'ems_id': self._ems_id,
                'id': "[-hub-][entity-type-group][[--][internal-type-group][root]]",
                'name': "<root>",
                'nodetype': "root",
                'parent_id': None
            }
            self._trees['dbtree'] = pd.DataFrame([dbroot])
            self.__update_children(dbroot, treetype="dbtree")
            self.update_tree("fdw", treetype="dbtree", exclude_tree=["APM Events"])
            self.__save_dbtree()
            tree = self._trees['dbtree']
        return tree

    def __save_dbtree(self):
        if len(self._trees['dbtree']) > 0:
            self._metadata.delete_data("dbtree", self.__get_filter('dbtree'))
            self._metadata.append_data("dbtree", self._trees['dbtree'])

    def __get_kvmaps(self):
        tree = self._metadata.get_data("kvmaps", self.__get_filter('kvmaps'))
        return tree

    def __save_kvmaps(self):
        if len(self._trees['kvmaps']) > 0:
            self._metadata.delete_data("kvmaps", self.__get_filter('kvmaps'))
            self._metadata.append_data("kvmaps", self._trees['kvmaps'])

    def set_database(self, name):
        """
        Method to select a database from the available databases in dbtree

        Parameters
        ----------
        name: str
            database name

        Returns
        -------
        None
        """
        tree = self._trees['dbtree']
        if self.searchtype == 'contain':
            dbs = tree[(tree.nodetype == 'database')
                       & tree.name.str.contains(_treat_spchar(name), case=False)]
            # Return database with shortest name
            self._db_id = dbs.loc[dbs['name'].str.len().idxmin(), 'id']
        # Exact string matching
        elif self.searchtype == 'match':
            self._db_id = tree[(tree.nodetype == 'database') &
                tree.name.str.match(_treat_spchar(name), case=False)]['id'].values[0]
        self._trees['fieldtree'] = self.__get_fieldtree()
        if self._trees['fieldtree'].empty:
            self.__update_children(self.get_database(), treetype="fieldtree")
        print("Using database '%s'." % self.get_database()['name'])

    def get_database(self):
        """
        Method to get currently selected database from dbtree

        Returns
        -------
        dict
            dictionary for the selected database
        """
        tree = self._trees['dbtree']
        return tree[(tree.nodetype == "database") & (tree.id == self._db_id)].iloc[0].to_dict()

    def __db_request(self, parent):
        body = None
        if parent['nodetype'] == "database_group":
            body = {'groupId': parent['id']}
        resp_h, d = self._conn.request(
            uri_keys=('database', 'group'),
            uri_args=self._ems_id,
            body=body
        )
        d1 = []
        if len(d['databases']) > 0:
            d1 = [{'uri_root': parent['uri_root'] if 'uri_root' in parent.keys() else '',
                   'ems_id': parent['ems_id'],
                   'id': x['id'],
                   'nodetype': 'database',
                   'name': x['pluralName'],
                   'parent_id': parent['id']
                   } for x in d['databases']]
            for _d in d1:
                if _d['uri_root'] == '':
                    _d.pop('uri_root')
        d2 = []
        if len(d['groups']) > 0:
            d2 = [{'uri_root': parent['uri_root'] if 'uri_root' in parent.keys() else '',
                   'ems_id': parent['ems_id'],
                   'id': x['id'],
                   'nodetype': 'database_group',
                   'name': x['name'],
                   'parent_id': parent['id']} for x in d['groups']]
            for _d in d2:
                if _d['uri_root'] == '':
                    _d.pop('uri_root')
        return d1, d2

    def __fl_request(self, parent):
        body = None
        if parent['nodetype'] == "field_group":
            body = {'groupId': parent['id']}
        resp_h, d = self._conn.request(
            uri_keys=('database', 'field_group'),
            uri_args=(self._ems_id, self._db_id),
            body=body
        )
        d1 = []
        if len(d['fields']) > 0:
            d1 = [{'uri_root': parent['uri_root'] if 'uri_root' in parent.keys() else '',
                   'ems_id': parent['ems_id'],
                   'db_id': self._db_id,
                   'id': x['id'],
                   'nodetype': 'field',
                   'type': x['type'],
                   'name': x['name'],
                   'parent_id': parent['id']} for x in d['fields']]
            for _d in d1:
                if _d['uri_root'] == '':
                    _d.pop('uri_root')
        d2 = []
        if len(d['groups']) > 0:
            d2 = [{'uri_root': parent['uri_root'] if 'uri_root' in parent.keys() else '',
                   'ems_id': parent['ems_id'],
                   'db_id': self._db_id,
                   'id': x['id'],
                   'nodetype': 'field_group',
                   'type': None,
                   'name': x['name'],
                   'parent_id': parent['id']} for x in d['groups']]
            for _d in d2:
                if _d['uri_root'] == '':
                    _d.pop('uri_root')
        return d1, d2

    def __add_subtree(self, parent, exclude_tree=[], treetype='fieldtree', exclude_subtrees=False):
        print("On " + parent['name'] + "(" + parent['nodetype'] + ")" + "...")
        if treetype == "dbtree":
            searchtype = 'database'
            d1, d2 = self.__db_request(parent)
        else:
            searchtype = "field"
            d1, d2 = self.__fl_request(parent)

        if len(d1) > 0:
            self._trees[treetype] = self._trees[treetype].append(d1, ignore_index=True)
            plural = "s" if len(d1) > 1 else ""
            print("-- Added %d %s%s" % (len(d1), searchtype, plural))

        for x in d2:
            if exclude_subtrees:
                print("-- Excluded subtree: %s" % x['name'])
            else:
                self._trees[treetype] = self._trees[treetype].append(x, ignore_index=True)
                if len(exclude_tree) > 0:
                    if all([y not in x['name'] for y in exclude_tree]):
                        self.__add_subtree(x, exclude_tree, treetype)
                else:
                    self.__add_subtree(x, exclude_tree, treetype)

    def __get_children(self, parent_id, treetype='fieldtree'):
        tr = self._trees[treetype]
        # tr = tr[tr.nodetype != ('field' if treetype=='fieldtree' else 'dbtree')]
        if isinstance(parent_id, (list, tuple, pd.Series)):
            return tr[tr.parent_id.isin(parent_id)]
        return tr[tr.parent_id == parent_id]

    def __remove_subtree(self, parent, treetype='fieldtree'):
        tr = self._trees[treetype]
        chld = tr[tr.parent_id == parent['id']]

        # Update the instance tree by deleting children
        self._trees[treetype] = tr[tr.parent_id != parent['id']]

        # Iterate and do recursive removal of children of children
        leaftype = 'field' if treetype == 'fieldtree' else 'database'
        for i, x in chld[chld.nodetype != leaftype].iterrows():
            self.__remove_subtree(x, treetype=treetype)

    # def __remove_subtree(self, parent, rm_parent=True, treetype='fieldtree'):
    #     rm_list = list()
    #     if rm_parent:
    #         rm_list.append(parent['id'])
    #     parent_id = [parent['id']]
    #     cntr = 0
    #     while len(parent_id) > 0:
    #         child_id = self.__get_children(parent_id, treetype=treetype)['id'].tolist()
    #         rm_list += child_id
    #         parent_id = child_id
    #         cntr += 1
    #         if cntr > 1e4:
    #             sys.exit("Something's wrong. Subtree removal went over 10,000 iterations.")
    #     if len(rm_list) > 0:
    #         tr = self._trees[treetype]
    #         self._trees[treetype] = tr[~tr.id.isin(rm_list)]
    #         print("Removed the subtree of %s (%s) with total of %d nodes "
    #               "(fields/databases/groups)."
    #               % (parent['name'], parent['nodetype'], len(rm_list)))

    def __update_children(self, parent, treetype='fieldtree'):
        # This function updates the direct children of a parent node.
        print("On " + parent['name'] + "(" + parent['nodetype'] + ")" + "...")
        if treetype == "dbtree":
            searchtype = 'database'
            d1, d2 = self.__db_request(parent)
        else:
            searchtype = "field"
            d1, d2 = self.__fl_request(parent)

        tree = self._trees[treetype]
        self._trees[treetype] = tree[~((tree.nodetype == searchtype)
                                       & (tree.parent_id == parent['id']))]

        if len(d1) > 0:
            self._trees[treetype] = self._trees[treetype].append(d1, ignore_index=True)
            plural = "s" if len(d1) > 1 else ""
            print("-- Added %d %s%s" % (len(d1), searchtype, plural))

        # If there is an array of groups as children add any that appeared new
        # and remove who does not.
        old_groups = tree[(tree.nodetype == '%s_group' % searchtype)
                          & (tree.parent_id == parent['id'])]
        old_ones = old_groups["id"].tolist()
        new_ones = [x['id'] for x in d2]

        rm_id = _listdiff(old_ones, new_ones)
        if len(rm_id) > 0:
            [self.__remove_subtree(x.to_dict(), treetype=treetype)
             for i, x in old_groups.iterrows() if x['id'] in rm_id]

        add_id = _listdiff(new_ones, old_ones)
        if len(add_id) > 0:
            self._trees[treetype] =\
                self._trees[treetype].append([x for x in d2 if x['id'] in add_id])

    def update_tree(self, *args, **kwargs):
        """
        Parameters
        ----------
        args:
            paths
        kwargs:
            keyword arguments

        Returns
        -------
        None

        Keyword arguments
        -----------------
        treetype: str
            "fieldtree" or "dbtree"
        exclude_tree: list
            Exact name strings (case sensitive) of the field groups you don't want to search through
            Ex. ['Profiles', 'Weather Information']
        exclude_subtrees: bool
            whether to exclude all subtrees or not
        """
        treetype = kwargs.get("treetype", "fieldtree")
        exclude_tree = kwargs.get("exclude_tree", [])
        exclude_subtrees = kwargs.get("exclude_subtrees", False)
        searchtype = "field" if treetype == "fieldtree" else "database"

        if treetype not in ("fieldtree", "dbtree"):
            raise ValueError("treetype = '%s': there is no such data table." % treetype)

        fld_path = [s.lower() for s in args]

        for idx, path in enumerate(fld_path):
            path = _treat_spchar(path)
            if idx == 0:
                tree = self._trees[treetype]
                if self.searchtype == 'contain':
                    parent = tree[tree.name.str.contains(path, case=False)]
                elif self.searchtype == 'match':
                    parent = tree[tree.name.str.match(path, case=False)]
            else:
                self.__update_children(parent, treetype=treetype)
                chld_df = self.__get_children(parent['id'], treetype=treetype)
                if self.searchtype == 'contain':
                    child = chld_df[chld_df.name.str.contains(path, case=False)]
                elif self.searchtype == 'match':
                    child = chld_df[chld_df.name.str.match(path, case=False)]
                parent = child
            if len(parent) == 0:
                raise ValueError("Search keyword '%s' did not return any %s group."
                                 % (path, searchtype))
            ptype = "%s_group" % searchtype
            parent = parent[parent.nodetype == ptype]
            parent = _get_shortest(parent)

        print("=== Starting to update subtree from '%s (%s)' ==="
              % (parent['name'], parent['nodetype']))
        self.__remove_subtree(parent, treetype=treetype)
        self.__add_subtree(
            parent,
            exclude_tree,
            treetype=treetype,
            exclude_subtrees=exclude_subtrees
        )

    def make_default_tree(self):
        """
        Method to create a default tree

        Returns
        -------
        None
        """
        dbnode = self.get_database()
        self.__remove_subtree(dbnode, treetype="fieldtree")
        self.__add_subtree(
            self.get_database(),
            exclude_tree=Flight.temp_exclude,
            treetype='fieldtree'
        )
        # self.update_tree(
        #     self.get_database()['name'],
        #     exclude_tree=Flight.temp_exclude,
        #     treetype='fieldtree'
        # )

    def search_fields(self, *args, **kwargs):
        """
        This function searches through the field names and returns a list of tuples with
        (field_name, field_info). The input arg is a "keyword" for the field name. You can
        also add parent field group names as n-tuple.

        Examples
        --------
        search_fields("Takeoff Airport Code", "Landing Airport Code")

        search_fields(("Takeoff","Airport","Code"), ("Landing","Airport","Airport Code"))

        Parameters
        ----------
        args:
            fields
        kwargs:
            keyword arguments

        Keyword arguments
        -----------------
        unique: bool
            Indicates if the value with the shortest name will be returned

        Returns
        -------
        res: list
            list of dictionaries
        """
        unique = kwargs.get("unique", True)
        res = pd.DataFrame()
        for field in args:
            if isinstance(field, tuple) and len(field) > 1:
                # If the given keyword is a tuple, search through the tree
                child = self._trees['fieldtree']
                for idx, ff in enumerate(field):
                    ff = _treat_spchar(ff)
                    if idx < (len(field)-1):
                        child = child[child.nodetype == "field_group"]
                        if self.searchtype == 'contain':
                            parent_id = child[child.name.str.contains(ff, case=False)]['id'].tolist()
                        elif self.searchtype == 'match':
                            parent_id = child[child.name.str.match(ff, case=False)]['id'].tolist()
                        tr = self._trees['fieldtree']
                        child = tr[tr.parent_id.isin(parent_id)]
                    else:
                        if self.searchtype == 'contain':
                            child = child[(child.nodetype == "field")
                                          & child.name.str.contains(ff, case=False)]
                        elif self.searchtype == 'match':
                            child = child[(child.nodetype == "field")
                                          & child.name.str.match(ff, case=False)]
                fres = child
            else:
                # Simple keyword search
                tree = self._trees['fieldtree']
                if self.searchtype == 'contain':
                    fres = tree[(tree.nodetype == "field")
                                & tree.name.str.contains(_treat_spchar(field), case=False)]
                elif self.searchtype == 'match':
                    fres = tree[(tree.nodetype == "field")
                                & tree.name.str.match(_treat_spchar(field), case=False)]

            if fres.shape[0] == 0:
                # No returned value. Raise error.
                raise ValueError("No field found with field keyword %s." % field)
            elif fres.shape[0] > 1:
                if unique:
                    # If more than one value returned, choose one with the shortest name.
                    fres = _get_shortest(fres)
            res = res.append(fres, ignore_index=True)

        # Convert the search result to a list of dicts
        res = [x[1].to_dict() for x in res.iterrows()]
        return res

    def list_allvalues(self, field=None, field_id=None, in_dict=False, in_df=False):
        """
        List all available values for a discrete field. Will raise error if the type of
        a given field is not discrete.

        Parameters
        ----------
        field: str
            field name
        field_id: str
            field id
        in_dict: bool
            returns output as a dictionary if set to True
        in_df: bool
            returns output as a pandas dataframe if set to True

        Returns
        -------
        dict | pd.DataFrame
            listed fields
        """
        if field_id is None:
            field = self.search_fields(field)[0]
            field_type = field['type']
            field_id = field['id']
            field_name = field['name']
            if field_type != 'discrete':
                sys.exit("Queried field should be discrete to get the list of possible values.")
        else:
            field_id = field_id
            tree = self._trees['fieldtree']
            field_name = tree.loc[tree.id == field_id, 'name'].values[0]

        tree = self._trees['kvmaps']
        kmap = tree[(tree.ems_id == self._ems_id) & (tree.id == field_id)]

        if len(kmap) == 0:
            print("%s: Getting key-value mappings from API. "
                  "(Caution: Some fields take a very long time)" % field_name)

            _, content = self._conn.request(
                uri_keys=('database', 'field'),
                uri_args=(self._ems_id, self._db_id, field_id)
            )
            km = content['discreteValues']
            kmap = pd.DataFrame({
                'uri_root': self._uri_root,
                'ems_id': self._ems_id,
                'id': field_id,
                'key': list(km.keys()),
                'value': list(km.values())
            })
            kmap['key'] = pd.to_numeric(kmap['key'])
            self._trees['kvmaps'] = self._trees['kvmaps'].append(kmap, ignore_index=True)
            self.__save_kvmaps()

        if in_dict:
            res = dict()
            for i, r in kmap.iterrows():
                res[r['key']] = r['value']
            return res

        if in_df:
            return kmap[['key', 'value']]
        return kmap['value'].tolist()

    def get_value_id(self, value, field=None, field_id=None):
        """
        Return the key (id) of the values of a discrete field.

        Parameters
        ----------
        value: str
            value name
        field: str
            field name
        field_id: str
            field id

        Returns
        -------
        int
            value id
        """
        kvmap = self.list_allvalues(field=field, field_id=field_id, in_df=True)
        key = kvmap[kvmap.value == value]['key']
        if len(key) == 0:
            raise ValueError("%s could not be found from the list of the field values." % value)
        return int(key.values[0])

    def __get_filter(self, tree):
        query_filter = "ems_id = %d" % self._ems_id
        if 'uri_root' in self._metadata.get_data(tree).columns:
            query_filter += " AND uri_root = '%s'" % self._uri_root
        return query_filter


def _get_shortest(fields):
    if isinstance(fields, pd.DataFrame) is False:
        sys.exit("Input should be a Pandas dataframe.")
    return fields.loc[fields.name.str.len().idxmin()].to_dict()


def _listdiff(a, b):
    return [x for x in a if x not in b]


def _treat_spchar(s):
    sp_chr = (
        ".^()[]{}<>-+?!*$|&%"
    )
    for x in sp_chr:
        s = s.replace(x, "\\"+x)
    return s



from __future__ import absolute_import
from __future__ import print_function

import json
import sys
from builtins import str
from builtins import zip
from collections import OrderedDict

import pandas as pd
from future.utils import string_types

from emspy.query import *
from .query import Query


class FltQuery(Query):
    """
    Flight query class
    """
    def __init__(self, conn, ems_name=None, data_file=LocalData.default_data_file):
        """
        Flight query initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        ems_name: str
            EMS system name
        data_file: str
            path to database file
        """
        Query.__init__(self, conn, ems_name)
        self._init_assets(data_file)
        self.reset()

    def _init_assets(self, data_file):
        # Query._init_assets(self)
        self.__flight = Flight(self._conn, self._ems_id, data_file)

    def set_database(self, name):
        """
        Set database for querying

        Parameters
        ----------
        name: str
            database name

        Returns
        -------
        None
        """
        self.__flight.set_database(name)

    def get_database(self):
        """
        Get selected database

        Returns
        -------
        dict
            dictionary for the selected database
        """
        return self.__flight.get_database()

    def reset(self):
        """
        Resets the API query to send. This should be called before creating a new query.

        Returns
        -------
        None
        """
        self.__columns = []
        self.__queryset = {
            "select": [],
            "groupBy": [],
            "orderBy": [],
            "distinct": True,
            "format": "none"
        }

    def select(self, *args, **kwargs):
        """
        Functionally equivalent to SQL's select statement

        Parameters
        ----------
        args:
            fields to query
        kwargs:
            keyword arguments

        Keyword arguments
        -----------------
        aggregate: str
            aggregation to apply, one of:
            ['none', 'avg', 'count', 'max', 'min', 'stdev', 'sum', 'var']

        Returns
        -------
        None

        Examples
        --------
        Following is the example of select method to query three fields and one more with
        aggregation function applied. The values are appended until the whole query is
        reset.

        >>> query.select("customer id", "takeoff valid", "takeoff airport iata code")
        >>> query.select("p301: fuel burned by all engines during cruise", aggregate="avg")
        """
        aggs = ['none', 'avg', 'count', 'max', 'min', 'stdev', 'sum', 'var']
        aggregate = kwargs.get('aggregate', 'none')
        if aggregate not in aggs:
            sys.exit("Wrong aggregation selected. Use one of %s." % aggs)
        fields = self.__flight.search_fields(*args, **kwargs)

        if not isinstance(fields, list):
            fields = [fields]

        for field in fields:
            d = {
                'fieldId': field['id'],
                'aggregate': aggregate
            }
            self.__queryset['select'].append(d)
            self.__columns.append(field)

    def deselect(self, *args):
        """
        Removes fields from the query

        Parameters
        ----------
        args:
            fields to remove

        Returns
        -------
        None
        """
        fields = self.__flight.search_fields(*args)
        if not isinstance(fields, list):
            fields = [fields]
        for field in fields:
            matching_entries = [d for d in self.__queryset['select'] if d['fieldId'] == field['id']]
            for entry in matching_entries:
                self.__queryset['select'].remove(entry)
            self.__columns.remove(field)

    def group_by(self, *args):
        """
        Functionally equivalent to SQL's groupby

        Parameters
        ----------
        args:
            fields

        Returns
        -------
        None
        """
        for field in self.__flight.search_fields(*args):
            self.__queryset['groupBy'].append({'fieldId': field['id']})

    def order_by(self, field, order='asc'):
        """
        Functional equivalent of SQL's order by

        Parameters
        ----------
        field: str
            field
        order: str
            ordering order:
                'asc': ascending
                'desc': descending

        Returns
        -------
        None
        """
        if order not in ['asc', 'desc']:
            sys.exit("Ordering option must be one of %s." % ['asc', 'desc'])
        self.__queryset['orderBy'].append({
            'fieldId': self.__flight.search_fields(field)[0]['id'],
            'order': order,
            'aggregate': 'none'})

    def filter(self, expr, operator='and'):
        """
        Translate the give filtering conditions to json queries.

        Parameters
        ----------
        expr: str
            filtering expression
        operator: str
            specifies how to aggregate filters:
                and: all filter conditions must be met
                or: any filter condition must be met

        Returns
        -------
        None
        """
        if 'filter' not in self.__queryset:
            self.__queryset['filter'] = dict(operator=operator, args=[])
        expr_vec = self.__split_expr(expr)
        jsonobj = self.__translate_expr(expr_vec)
        self.__queryset['filter']['args'].append(jsonobj)

    def remove_filter(self, expr):
        """
        Removes a filter from the queryset

        Parameters
        ----------
        expr: str
            filtering expression

        Returns
        -------
        None
        """
        if 'filter' in self.__queryset:
            expr_vec = self.__split_expr(expr)
            jsonobj = self.__translate_expr(expr_vec)
            # Remove the filter from the arguments
            self.__queryset['filter']['args'].remove(jsonobj)
            # If no filters are left remove the filter key from the queryset
            if len(self.__queryset['filter']['args']) == 0:
                self.__queryset.remove('filter')

    @staticmethod
    def __split_expr(expr):
        # This function might need to be updated to be more robust. The split of a
        # given expression will not be correctly done if either field or field
        # value contains symbols identical to filtering operators.
        #
        # In fact the correction was already applied in Rems so refer to Rems'
        # split_expr function to check what was done there.
        import re

        any_valid_match = 0
        for pattern in ['[=!<>]=?'] + list(sp_ops.keys()):
            # The regular expression looks for either of these patterns in the filter:
            # 1. <variable> <operator> <value> (i.e. "'x' < '1'")
            # 2. <value> <operator> <variable> <operator> <value> (i.e. "'0' < 'x' < '1'")
            # Pattern 1 is the standard for filtering
            # Pattern 2 is only used for 'between' filtering
            if pattern not in ['is null', 'is not null']:
                match = re.search(
                    r"(.*)\s+(%s)\s+(.*)\s+(%s)\s+(.*)|(.*)\s+(%s)\s+(.*)".replace('%s', pattern),
                    expr
                )
            else:
                match = re.search(r"(.*)\s+(%s)$".replace('%s', pattern), expr)
            if match is not None:
                any_valid_match = len([group for group in match.groups() if group is not None])
                if any_valid_match:
                    break

        if match is None or not any_valid_match:
            sys.exit("Cannot find any valid conditional operator from the given string expression.")
        split = [group.strip() for group in match.groups() if group is not None]
        return split

    def __translate_expr(self, expr_vec):
        fld_loc = [False, False]
        fld_info = None
        fld_type = None
        val_info = []

        # If the expression vector contains 3 elements the filter is in the form
        # <variable> <operator> <value>
        # The operator is the second item of the list
        if len(expr_vec) in [2, 3]:
            op = expr_vec[1]
        # If the expression vector contains 5 elements the filter is in the form
        # <value> <operator> <variable> <operator> <value>
        # The operators are the second and fourth elements of the list
        elif len(expr_vec) == 5:
            op = [expr_vec[1], expr_vec[3]]

        # Get the list of operators in the expression
        op_vec = [op] if not isinstance(op, list) else op
        # Remove the operator[s] from the expression vector
        fld_val_vec = [s for s in expr_vec if s not in op_vec]
        for i, s in enumerate(fld_val_vec):
            x = eval(s)  # FIXME Remove eval
            # If the length of the expression vector is 3, the field is the first item in the list.
            # Otherwise, if the length of the expression vector is 5, the second item in the list
            # is the field.
            if i == 0 if len(expr_vec) in [2, 3] else i == 1:
                fld = self.__flight.search_fields(x)[0]
                if fld is not None:
                    fld_info = [{'type': 'field', 'value': fld['id']}]
                    fld_type = fld['type']
                    fld_loc[i] = True
                else:
                    raise ValueError("No field was found with the keyword %s. "
                                     "Please double-check if it is a right keyword." % x)
            else:
                if type(x) != list:
                    x = [x]
                val_info += [{'type': 'constant', 'value': v} for v in x]

        # This logic only applies to standard conditional operators
        if fld_loc[1] and len(expr_vec) == 3:
            if '<' in expr_vec[1]:
                op = expr_vec[1].replace('<', '>')
            else:
                op = expr_vec[1].replace('>', '<')

        arg_list = fld_info + val_info

        if fld_type == "boolean":
            fltr = _boolean_filter(op, arg_list)
        elif fld_type == "discrete":
            fltr = _discrete_filter(op, arg_list, self.__flight)
        elif fld_type == "number":
            fltr = _number_filter(op, arg_list)
        elif fld_type == "string":
            fltr = _string_filter(op, arg_list)
        elif fld_type == "dateTime":
            fltr = _datetime_filter(op, arg_list)
        else:
            raise ValueError("%s has an unknown field data type %s." % (fld[0], fld_type))
        return fltr

    def distinct(self, x=True):
        """
        Functionally equivalent to SQL's distinct

        Parameters
        ----------
        x: bool
            sets distinct flag if True

        Returns
        -------
        None
        """
        self.__queryset['distinct'] = x

    def get_top(self, n):
        """
        Functionally equivalent to SQL's limit or top
        Parameters
        ----------
        n: int
            number of records to limit the query to

        Returns
        -------
        None
        """
        self.__queryset['top'] = n

    # def readable_output(self, x=False):
    #     if x:
    #         y = "display"
    #     else:
    #         y = "none"
    #     self.__queryset['format'] = y

    def from_json_string(self, json_str):
        """
        Load the JSON string for this query.

        Parameters
        ----------
        json_str: str
            Queryset as a JSON str.
        """

        if type(json_str) is not str:
            raise TypeError('json_str should be a str. Found: {0}'.format(type(json_str)))

        self.__queryset = json.loads(json_str)

    def in_json(self):
        """
        Dump the queryset as JSON

        Returns
        -------
        str
            queryset as JSON string
        """
        return json.dumps(self.__queryset)

    def in_dict(self):
        """
        Returns the queryset as dictionary

        Returns
        -------
        dict
            queryset as dictionary
        """
        return self.__queryset

    def simple_run(self, output="dataframe"):
        """
        Sends query to EMS API via the regular query call. The regular query call has a size limit
        in the returned data, which is 25000 rows max. Any output that has greater than 25000 rows
        will be truncated. For the query that is expected to return with large data. Please use the
        async_run method.

        Parameters
        ----------
        output: str
            desired output data format. Either "raw" or "dataframe".

        Returns
        -------
        pd.Dataframe
            Returned data for query in Pandas' DataFrame format
        """
        print('Sending a simple query to EMS ...')
        resp_h, content = self._conn.request(
            rtype="POST",
            uri_keys=('database', 'query'),
            uri_args=(self._ems_id, self.__flight.get_database()['id']),
            jsondata=self.__queryset
        )
        print('Done.')

        if output == "raw":
            return content
        elif output == "dataframe":
            return self.__to_dataframe(content)
        else:
            raise ValueError("Requested an unknown output type.")

    def async_run(self, n_row=25000):
        """
        Sends query to EMS API via async-query call. The async-query does not process
        the query as a single batch for a query expecting a large data. You will have
        to call it multiple times. This function do this multiple calls for you.

        Parameters
        ----------
        n_row: int
            batch size of a single async call. Default is 25000.

        Returns
        -------
        pd.DataFrame
            Returned data for query in Pandas' DataFrame format
        """
        print('Sending and opening an async-query to EMS ...', end=' ')
        resp_h, content = self._conn.request(
            rtype="POST",
            uri_keys=('database', 'open_asyncq'),
            uri_args=(self._ems_id, self.__flight.get_database()['id']),
            jsondata=self.__queryset
        )
        if 'id' not in content:
            sys.exit("Opening Async query did not return the query Id.")
        query_id = content['id']
        query_header = content['header']
        print('Done.')

        ctr = 0
        df = None
        while True:
            print(" === Async call: %d ===" % (ctr+1))
            try:
                resp_h, content = self._conn.request(
                    rtype="GET",
                    uri_keys=('database', 'get_asyncq'),
                    uri_args=(
                        self._ems_id,
                        self.__flight.get_database()['id'],
                        query_id,
                        n_row * ctr,
                        n_row * (ctr+1) - 1)
                    )
                content['header'] = query_header
                dff = self.__to_dataframe(content)
            except:
                print("Something's wrong. Returning what has been sent so far.")
                # from pprint import pprint
                # pprint(resp_h)
                # pprint(content)
                return df

            if ctr == 0:
                df = dff
            else:
                df = df.append(dff, ignore_index=True)

            print("Received up to %d rows." % df.shape[0])
            if dff.shape[0] < n_row:
                break
            ctr += 1

        print("Done.")
        return df

    def run(self, n_row=25000):
        """
        Sends query to EMS API. It uses either regular or async query call depending on
        the expected size of output data. It supports only Pandas DataFrame as the output
        format.

        Parameters
        ----------
        n_row: int
            batch size of a single async call. Default is 25000.

        Returns
        -------
        pd.DataFrame
            Returned data for query in Pandas' DataFrame format
        """
        Nout = None
        if 'top' in self.__queryset:
            Nout = self.__queryset['top']

        if (Nout is not None) and (Nout <= 25000):
            return self.simple_run(output="dataframe")

        return self.async_run(n_row=n_row)

    def __to_dataframe(self, json_output):
        # Changes Dict (JSON) formatted raw output from the EMS API to Pandas' DataFrame.
        print("Raw JSON output to Pandas dataframe...")
        col = [h['name'] for h in json_output['header']]
        coltypes = [c['type'] for c in self.__columns]
        col_id = [c['id'] for c in self.__columns]
        val = json_output['rows']

        df = pd.DataFrame(data=val, columns=col)

        if df.empty:
            return df

        if self.__queryset['format'] == "display":
            print("Done.")
            return df

        # Do the dirty work of casting a right type for each column of the data
        # Note
        # ====
        # Runway IDs are discrete data but their key-value mapping is not provided
        # because the mapping itself is quite big in size (45K entries). That means
        # the regular routines to handle the discrete data won't work. As a result
        # the discrete data routine has a dirty, custom routine particularly for
        # the runway IDs. What it basically does is to send a separate but redundant
        # query for runway IDs with "queryset$format = display", and then push the
        # this query result at the runway ID column of the original query result.
        # I know this is crappy but it seems the best way I could find.
        for i, cid, cname, ctype in zip(range(len(col)), col_id, col, coltypes):
            try:
                if ctype == 'number':
                    df.iloc[:, i] = pd.to_numeric(df.iloc[:, i])
                elif ctype == 'discrete':
                    df.iloc[:, i] = self.__key_to_val(df.iloc[:, i], cid)
                    # k_map = self.__flight.list_allvalues(field_id=cid, in_dict=True)
                    # if len(k_map) == 0:
                    #     df[cname] = self.__get_rwy_id(cname)
                    # else:
                    #     df = df.replace({cname: k_map})
                elif ctype == 'boolean':
                    df.iloc[:, i] = df.iloc[:, i].astype(bool)
                elif ctype == 'dateTime':
                    df.iloc[:, i] = pd.to_datetime(df.iloc[:, i], utc=True)
            except ValueError:
                print("Somethings wrong when converting to Pandas DataFrame for column '%s' "
                      "(type: %s)." % (cname, ctype))
        print("Done.")
        return df

    def __key_to_val(self, ds, field_id):
        k_map = self.__flight.list_allvalues(field_id=field_id, in_df=True)
        # Sometimes k_map is a very large table making "replace" operation
        # very slow. Just grap kv-maps subset that are present in the target
        # dataframe
        k_map = k_map[k_map.key.isin(ds.unique())]
        # Change k_map in dict
        km_dict = dict()
        for i, r in k_map.iterrows():
            km_dict[r['key']] = r['value']
        ds = ds.replace(km_dict)
        return ds

    def __get_rwy_id(self, cname):
        # Deprecated
        print("\n --Running a special routine for querying runway IDs. "
              "This will make the querying twice longer.")
        qs = self.__queryset
        qs['format'] = "display"

        resp_h, content = self._conn.request(
            rtype="POST",
            uri_keys=('database', 'query'),
            uri_args=(self._ems_id, self.__flight.get_database()['id']),
            jsondata=qs
        )
        return self.__to_dataframe(content)[cname]

    def update_dbtree(self, *args, **kwargs):
        """
        Update database tree

        Parameters
        ----------
        args:
            database paths
        kwargs
            keyword arguments

        Keyword arguments
        -----------------
        exclude_tree: list
            trees to exclude from database tree

        Returns
        -------
        None
        """
        exclude_tree = kwargs.get("exclude_tree", [])
        self.__flight.update_tree(*args, **{"treetype": "dbtree", "exclude_tree": exclude_tree})

    def update_fieldtree(self, *args, **kwargs):
        """
        Update field tree

        Parameters
        ----------
        args:
            field paths
        kwargs
            keyword arguments

        Keyword arguments
        -----------------
        exclude_tree: list
            trees to exclude from database tree
        exclude_subtrees: bool
            excludes all subtrees if set to True

        Returns
        -------
        None
        """
        exclude_tree = kwargs.get("exclude_tree", [])
        exclude_subtrees = kwargs.get("exclude_subtrees", False)
        self.__flight.update_tree(*args, **{"treetype": "fieldtree",
                                            "exclude_tree": exclude_tree,
                                            "exclude_subtrees": exclude_subtrees})

    def generate_preset_fieldtree(self):
        """
        Creates a default field tree

        Returns
        -------
        None
        """
        self.__flight.make_default_tree()

    def save_metadata(self, file_name=None):
        """
        Saves metadata to the set path

        Parameters
        ----------
        file_name: str
            output path

        Returns
        -------
        None
        """
        self.__flight.save_tree(file_name)

    def load_metadata(self, file_name=None):
        """
        Loads metadata from set path

        Parameters
        ----------
        file_name: str
            input path

        Returns
        -------
        None
        """
        self.__flight.load_tree(file_name)


basic_ops = {
    '==': 'equal',
    '!=': 'notEqual',
    '<': 'lessThan',
    '<=': 'lessThanOrEqual',
    '>': 'greaterThan',
    '>=': 'greaterThanOrEqual'
}
between_ops = {
    '<': {
        '<': 'betweenExclusive'
    },
    '<=': {
        '<=': 'betweenInclusive'
    },
    '>': {
        '>': 'notBetweenExclusive'
    },
    '>=': {
        '>=': 'notBetweenInclusive'
    }
}
sp_ops = OrderedDict([
    ('not in', 'notIn'),
    ('in', 'in'),
    ('is null', 'isNull'),
    ('is not null', 'isNotNull')
])

# '=Null': 'isNull', '!=Null': 'isNotNull', 'and': 'And', 'or': 'Or', 'in': 'in', 'not in': 'notIn'


def _filter_fmt1(op, *args):
    fltr = {
        "type": "filter",
        "value": {
            "operator": op,
            "args": []
        }
    }
    for x in args:
        fltr['value']['args'].append({'type': x['type'], 'value': x['value']})
    return fltr


def _boolean_filter(op, d):
    # Between filters unsupported
    if len(d) > 2:
        raise ValueError('Unsupported conditional operator for boolean field.')
    fld_info = d[0]
    if len(d) == 1:
        if op.strip().lower() == 'is null':
            t_op = 'isNull'
        elif op.strip().lower() == 'is not null':
            t_op = 'isNotNull'
    elif len(d) == 2:
        val_info = d[1]
        if not isinstance(val_info['value'], bool):
            raise ValueError("%s: use a boolean value." % val_info['value'])
        if op == "==":
            t_op = 'is'+str(val_info['value'])
        elif op == "!=":
            t_op = 'is'+str(not val_info['value'])
        else:
            raise ValueError("Conditional operator %s is given. "
                             "Booleans should be only with boolean operators." % op)
    fltr = _filter_fmt1(t_op, fld_info)
    return fltr


def _discrete_filter(op, d, flt):
    fld_info = d[0]

    # Between filters unsupported
    if isinstance(op, list):
        raise ValueError("%s: Unsupported conditional operators for discrete field type." % op)

    if op in list(basic_ops.keys()):
        # Single input basic conditional operation
        t_op = basic_ops[op]
        val_info = d[1]
        vid = flt.get_value_id(val_info['value'], field_id=fld_info['value'])
        val_info['value'] = vid
        fltr = _filter_fmt1(t_op, fld_info, val_info)
    elif op.strip() in sp_ops.keys():
        t_op = sp_ops[op]
        val_list = [
            {'type': x['type'], 'value': flt.get_value_id(x['value'], field_id=fld_info['value'])}
            for x in d[1:]
        ]
        inp = [fld_info] + val_list
        fltr = _filter_fmt1(t_op, *inp)
    else:
        raise ValueError("%s: Unsupported conditional operator for discrete field type." % op)
    return fltr


def _number_filter(op, d):
    if isinstance(op, string_types):
        if op in basic_ops.keys():
            t_op = basic_ops[op]
            fltr = _filter_fmt1(t_op, d[0], d[1])
        elif op in sp_ops.keys():
            t_op = sp_ops[op]
            fltr = _filter_fmt1(t_op, *d)
        else:
            raise ValueError("%s: Unsupported conditional operator for number field type." % op)
    elif isinstance(op, list):
        if op[1] in between_ops[op[0]]:
            t_op = between_ops[op[0]][op[1]]
            fltr = _filter_fmt1(t_op, d[0], d[1], d[2])
        else:
            raise ValueError("%s: Unsupported conditional operators for number field type." % op)
    return fltr


def _string_filter(op, d):
    # Between filters unsupported
    if isinstance(op, list):
        raise ValueError("%s: Unsupported conditional operators for string field type." % op)

    if op in ["==", "!="]:
        t_op = basic_ops[op]
        fltr = _filter_fmt1(t_op, d[0], d[1])
    elif op.strip() in sp_ops.keys():
        t_op = sp_ops[op]
        fltr = _filter_fmt1(t_op, *d)
    else:
        raise ValueError("%s: Unsupported conditional operator for string field type." % op)

    return fltr


def _datetime_filter(op, d):
    basic_date_ops = {
        "<": "dateTimeBefore",
        ">=": "dateTimeOnAfter",
    }
    sp_date_ops = {
        "is null": "isNull",
        "is not null": "isNotNull"
    }

    # Between filters unsupported
    if isinstance(op, list):
        raise ValueError("%s: Unsupported conditional operators for datetime field type." % op)

    if op in list(basic_date_ops.keys()):
        t_op = basic_date_ops[op]
        fltr = _filter_fmt1(t_op, *d)
        # Additional json attribute to specify this is UTC time
        fltr['value']['args'].append({'type': 'constant', 'value': 'Utc'})
    elif op in list(sp_date_ops.keys()):
        t_op = sp_date_ops[op]
        fltr = _filter_fmt1(t_op, *d)
    else:
        raise ValueError("%s: Unsupported conditional operator for datetime field type." % op)

    return fltr



from __future__ import print_function
from builtins import range
from builtins import object
import pandas as pd
import numpy as np
import sys
from emspy.query import LocalData


class Analytic(object):
	"""
	Analytic class
	"""
	def __init__(self, conn, ems_id, data_file=LocalData.default_data_file, searchtype='contain'):
		"""
		Analytic class initialization

		Parameters
		----------
		conn: emspy.connection.Connection
			connection object
		ems_id: int
			EMS system id
		data_file: str
			path to local DB file
		searchtype: str
			search type to perform:
				contain: uses the Pandas string method 'contains'
				match: uses the Pandas string method 'match'
		"""
		self._conn = conn
		self._ems_id = ems_id
		self._metadata = None
		self._load_paramtable(data_file)
		self.searchtype = searchtype

	def _load_paramtable(self, file_name=LocalData.default_data_file):
		# Load parameter table
		if self._metadata is None:
			self._metadata = LocalData(file_name)
		elif not self._metadata.is_db_path_correct(file_name):
			self._metadata.close()
			self._metadata = LocalData(file_name)

		self._param_table = self._metadata.get_data("params", "ems_id = %d" % self._ems_id)

	def _save_paramtable(self):
		# Save parameter table
		if len(self._param_table) > 0:
			self._metadata.delete_data("params", "ems_id = %d" % self._ems_id)
			self._metadata.append_data("params", self._param_table)

	def get_param_details(self, analytic_id):
		"""
		Method to get parameter details from the API

		Parameters
		----------
		analytic_id: str
			unique analytic id

		Returns
		-------
		content: list
			response from API
		"""
		_, content = self._conn.request(
			rtype="POST",
			uri_keys=('analytic', 'search'),
			uri_args=self._ems_id,
			jsondata={'id': analytic_id}
		)
		return content

	def search_param(self, keyword, in_df=False):
		"""
		Method to search a parameter in EMS and return the following fields:
			- id: EMS id
			- name: parameter name
			- description: parameter description
			- units: parameter units
			- ems_id: ems system id

		Parameters
		----------
		keyword: str
			parameter to search
		in_df: bool
			selector for output type, if True returns a dataframe,
			if False returns a list
		Returns
		-------
		pd.DataFrame or list
			parameter information (id, name, description, units, ems_id)
		"""
		print('Searching for params with keyword "%s" from EMS ...' % keyword, end=' ')
		# EMS API Call
		_, content = self._conn.request(
			uri_keys=('analytic', 'search'),
			uri_args=self._ems_id,
			body={'text': keyword}
		)
		if len(content) == 0:
			sys.exit("No parameter found with search keyword %s." % keyword)
		elif len(content) == 1:
			res = content
		else:
			word_len = [len(x['name']) for x in content]
			idx = np.argsort(word_len).tolist()
			res = [content[i] for i in idx]
		print("done.")

		for i in range(len(res)):
			res[i]['ems_id'] = self._ems_id

		if in_df:
			return pd.DataFrame(res)
		
		return res

	def get_param(self, keyword, unique=True):
		"""
		Get parameter information

		Parameters
		----------
		keyword: str
			parameter to fetch
		unique: bool
			indicates whether to return the parameter with the shortest name (if True)

		Returns
		-------
		dict
			fetched parameter information (uri_root, ems_id, id, name, description, units)
		"""
		# if the param table is empty, just return an empty param dict.
		if self._param_table.empty:
			return dict(ems_id="", id="", name="", description="", units="")
		# If the param table is not empty, do search by keyword
		if self.searchtype == 'contain':
			bool_idx = self._param_table['name'].str.contains(keyword, case=False, regex=False)
		elif self.searchtype == 'match':
			bool_idx = self._param_table['name'].str.match(keyword, case=False)
		df = self._param_table[bool_idx]
		# If the search result is empty, return empty param dict
		if df.empty:
			return dict(ems_id="", id="", name="", description="", units="")
		# If not empty, return the one with shortest name
		if df.shape[0] > 1:
			idx = df['name'].map(lambda x: len(x)).sort_values().index
			df = df.loc[idx, :]
		# When unique = True
		if unique:
			return df.iloc[0, :].to_dict()
		# When unique = False
		return df.to_dict('records')

	def get_physical_parameter_list(self, flight_id):
		"""
		Retrieves the physical parameters available for a certain `flight_id`

		Parameters
		----------
		flight_id: int
			Flight ID to pull physical parameters for.

		Returns
		-------
		pd.DataFrame
			A Pandas DataFrame with the columns: `id`, `name`, `description`, `units`
		"""

		_, content = self._conn.request(
			uri_keys=('analytic', 'group_f'),
			uri_args=(self._ems_id, flight_id),
			body={"analyticGroupId": "RG.RawParameters"}
		)

		content_analytics = content['analytics']
		df = pd.DataFrame(content_analytics)

		return df

	def get_flight_analytic_metadata(self, analytic_id, flight_id):
		"""
		Retrieves the analytic metadata in the context of a flight. This is especially useful when you are querying
		physical parameters, which only make sense in the context of a given flight (as opposed to globally). There
		will be one key:value pair for every piece of metadata returned by the API. Keys could vary depending on the
		type of analytic.

		Parameters
		----------
		analytic_id: str
			Analytic ID to pull metadata for.
		flight_id: str
			Example Flight ID to pull metadata for.

		Returns
		-------
		dict
			A dict with metadata fields returned by the API. There will be one key:value pair for every piece of
			metadata returned by the API. Keys could vary depending on the type of analytic.
		"""

		_, content = self._conn.request(
			uri_keys=('analytic', 'metadata'),
			uri_args=(self._ems_id, flight_id),
			jsondata={"id": analytic_id}
		)

		api_return_values = content['values']

		# Values are returned as {"key": "key_name", "value": "value_name"}, so we can simply unpack those into a more
		# user-friendly dictionary.
		return_dict = {item['key']: item['value'] for item in api_return_values}

		return return_dict



from __future__ import absolute_import
from .asset import Asset


class Fleet(Asset):
	"""
	Manages fleet-wise querying
	"""

	def __init__(self, conn, ems_id):
		"""
		Fleet asset object initialization

		Parameters
		----------
		conn: emspy.connection.Connection
			connection object
		ems_id: int
			EMS system id
		"""
		Asset.__init__(self, conn, "Fleet")
		self._ems_id = ems_id
		self.update_list()

	def update_list(self):
		"""
		Update fleet list

		Returns
		-------
		None
		"""
		Asset.update_list(self, uri_keys=('fleet', 'list'), uri_args=self._ems_id)
		self._rename_datacol('description', 'name')

	def get_id(self, name=None):
		"""
		Get fleet id from name

		Parameters
		----------
		name: str
			fleet name

		Returns
		-------
		fleet_id: int
			corresponding fleet name
		"""
		fleet_id = self.search('name', name)['id'].tolist()
		return fleet_id if len(fleet_id) > 1 else fleet_id[0]

	def get_name(self, id_val=None):
		"""
		Get fleet name from id

		Parameters
		----------
		id_val: int
			fleet id

		Returns
		-------
		fleet_name: str
			corresponding fleet name
		"""
		fleet_name = self.search('id', id_val, searchtype="match")['name'].tolist()
		return fleet_name if len(fleet_name) > 1 else fleet_name[0]



from .query import Query
from .ems import EMS
import pandas as pd
import json


class InsertQuery(Query):
    """
    A class for performing database inserts using the EMS API.
    """
    def __init__(self, conn, ems_name, database_id):
        """
        Query initialization

        Parameters
        ----------
        conn: ems.connection.Connection
            A valid instantiated EMS connection object.
        ems_name: str
            A valid EMS system name.
        database_id: str
            A database id to insert into.
        """
        Query.__init__(self, conn, ems_name)
        self._database_id = database_id
        self.__create_columns = {'createColumns': []}

    def in_json(self):
        """
        Returns data as JSON

        Returns
        -------
        str
            JSON serialization of the create_columns attribute
        """
        return json.dumps(self.__create_columns)

    def reset(self):
        """
        Resets the API query to send. This should be called before creating a new query.

        Returns
        -------
        None
        """
        self.__create_columns = {'createColumns': []}

    def get_create_columns(self):
        """
        A method which returns the values of createColumns attribute.

        Returns
        -------
        dict
            a dictionary corresponding to the current value of self.__create_columns
        """
        return self.__create_columns

    def insert_df(self, df, schema_map=None):
        """
        A method for inserting values from a pandas DataFrame.

        Parameters
        ----------
            df: pd.DataFrame
                A pandas DataFrame of values to input, where the columns are the fieldIds and the
                entries are values to input (unless schema_map is passed, in which case column names
                can be arbitrary as long as they exist in schema_map and map to ems monikers).
            schema_map: dict
                A mapping of named DataFrame columns to field ids, e.g.
                {'column1' = '[-hub][schema]'}. If this is not passed, the columns of df should
                correspond to EMS schemas.

        Raises
        ------
            TypeError
                If `schema_map` is not a dictionary.
            TypeError
                If `df` is not a pandas DataFrame.
            ValueError
                If `schema_map` is passed, but not all `df` column names are present in `schema_map`
        """
        if not isinstance(df, pd.DataFrame):
            raise TypeError("`df` must be a pandas DataFarme. Found: {0}".format(type(df)))

        # If a schema_map was passed in (mapping column name to ems schema)
        if schema_map is not None:
            # schema_map should be a dictionary.
            if not isinstance(schema_map, dict):
                raise TypeError("`schema_map` must be a dictionary.  Found: {0}".format(type(df)))
            # Find columns that exist in `df` but not in `schema_map`.  Convert to list.
            missing_from_map = list(set(df.columns.values) - set(schema_map.keys()))
            # If any columns are in `df` but not in `schema_map`, error out.
            if len(missing_from_map) > 0:
                raise ValueError("Column(s): '{0}' found in `df`, but not in `schema_map`."
                                 .format(missing_from_map))
            df = df.rename(columns=schema_map)
        else:
            print('Schema_map was not passed. '
                  'Assuming dataframe column names correspond to EMS schema.')

        for idx, row in df.iterrows():
            row_dict = row.to_dict()
            row_list = [
                {'fieldId': fid, 'value': val} for fid, val
                in zip(row_dict.keys(), row_dict.values())
            ]
            self.__create_columns['createColumns'].append(row_list)

    def run(self):
        """
        Method for posting the query to the API

        Returns
        -------
        None
        """
        _, content = self._conn.request(
            rtype='POST',
            uri_keys=('database', 'create'),
            uri_args=(self._ems_id, self._database_id),
            jsondata=self.__create_columns
        )
        # Unintended responses are handled in Connection.request, so we only do verification here.
        rows_added = content['rowsAdded']
        if rows_added == len(self.__create_columns['createColumns']):
            print('Added all rows.')



from __future__ import absolute_import
from .asset import Asset


class Airport(Asset):
	""" Manages airport info """

	_col_tr = {
		'default': 'codePreferred',
		'faa': 'codeFaa',
		'icao': 'codeIcao',
		'city': 'city',
		'name': 'name'
	}

	def __init__(self, conn, ems_id):
		"""
		Airport asset object initialization

		Parameters
		----------
		conn: emspy.connection.Connection
			connection object
		ems_id: int
			EMS system id
		"""
		Asset.__init__(self, conn, "Airport")
		self._ems_id = ems_id
		self.update_list()

	def update_list(self):
		"""
		Update airport list

		Returns
		-------
		None
		"""
		Asset.update_list(self, uri_keys=('airport', 'list'), uri_args=self._ems_id, colsort=False)

	def get_id(self, name=None, code='default'):
		"""
		Get airport id from ICAO code, FAA code, city or name

		Parameters
		----------
		name: str
			ICAO code, FAA code, city or name
		code: str
			input selector:
				- default: preferred code
				- faa: FAA code
				- icao: ICAO code
				- city: city
				- name: airport name

		Returns
		-------
		airport_id: int
			airport id
		"""
		airport_id = self.search(Airport._col_tr[code], name)['id'].tolist()
		return airport_id if len(airport_id) > 1 else airport_id[0]

	def get_name(self, id_val=None, code='default'):
		"""
		Get airport name from id

		Parameters
		----------
		id_val: int
			airport id
		code: str
			output selector:
				- default: preferred code
				- faa: FAA code
				- icao: ICAO code
				- city: city
				- name: airport name

		Returns
		-------
		airport_name: str
			airport output as per selector
		"""
		airport_name = self.search('id', id_val, searchtype="match")[Airport._col_tr[code]].tolist()
		return airport_name if len(airport_name) > 1 else airport_name[0]



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

from builtins import object
import gzip
import io
import json
import ssl
import sys
import urllib.error
import urllib.parse
import urllib.request

from numbers import Number
import pprint as pp
from . import common


class Connection(object):
    """
    Object for connection to EMS API
    """
    def __init__(self, user=None, pwd=None, proxies=None, verbose=False, ignore_ssl_errors=False,
                 server="prod", server_url=None, max_trials=3):
        """
        Connection initialization

        Parameters
        ----------
        user: str
            EMS username
        pwd: str
            EMS password
        proxies: dict
            proxies dictionary {'http': '', 'https': ''}
        verbose: bool
            verbose output (default False)
        ignore_ssl_errors: bool
            ignore SSL certificate errors (default False)
        server: str
            server to connect to from emspy.common.uri_root
        server_url: str
            hostname for server to connect to if different from any of the common values
        max_trials: int
            maximum number of reconnection attempts
        """
        self.__user = user
        self.__pwd = pwd
        self.__proxies = proxies
        self.__ntrials = 0
        self.__max_trials = max_trials
        self._uri_root = None
        self.__ignore_ssl_errors = ignore_ssl_errors
        self.token = None
        self.token_type = None

        # We assign the uri root to a member variable up front, and use that everywhere to
        # simplify. In order to use an alternate uri root, it must be specified in the constructor.
        if server_url is not None:
            if server_url.endswith('/api'):
                self._uri_root = server_url
            else:
                self._uri_root = server_url.rstrip('/') + '/api'
        else:
            self._uri_root = common.uri_root[server]

        if (user is not None) and (pwd is not None):
            self.connect(user, pwd, proxies, verbose)
        else:
            print("An empty connection is instantiated because credentials are not provided.\n")

    def connect(self, user, pwd, proxies=None, verbose=False):
        """
        Connect to EMS system using given credentials.

        Parameters
        ----------
        user: str
            EMS username
        pwd: str
            EMS password
        proxies: dict
            proxies dictionary (default None)
        verbose: bool
            verbose output (default False)

        Returns
        -------
        resp_h: list
            response headers
        content: dict
            response content
        """
        if proxies is not None:
            proxy_handler = urllib.request.ProxyHandler(proxies)
            opener = urllib.request.build_opener(proxy_handler, urllib.request.HTTPHandler)
            urllib.request.install_opener(opener)

        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'User-Agent': common.user_agent
        }
        data = {
            'grant_type': 'password',
            'username': user,
            'password': pwd
        }

        resp_h, content = self.request(
            rtype="POST",
            uri_keys=('sys', 'auth'),
            data=data,
            headers=headers,
            proxies=proxies,
            verbose=verbose
        )

        # Add error handling --

        # Get the token
        self.token = content['access_token']
        self.token_type = content['token_type']
        return resp_h, content

    def reconnect(self, verbose=False):
        """
        Method to attempt reconnection to the API

        Parameters
        ----------
        verbose: bool
            verbose output (default False)

        Returns
        -------
        resp_h: list
            response headers
        content: dict
            response content
        """
        if self.__ntrials >= self.__max_trials:
            sys.exit("Stop trying to reconnect EMS API after %d trials" % self.__ntrials)
        self.__ntrials += 1
        return self.connect(self.__user, self.__pwd, self.__proxies, verbose)

    def request(self, rtype="GET", uri=None, uri_keys=None, uri_args=None, headers=None,
                body=None, data=None, jsondata=None, proxies=None, verbose=False):
        """
        Method to send a HTTP request to the API

        Parameters
        ----------
        rtype: str
            request type (UNUSED)
        uri: str
            request uri
        uri_keys: tuple
            request uri keys
        uri_args: list, tuple
            request uri arguments
        headers: dict
            request headers
        body: dict
            request body
        data: dict
            request data
        jsondata:
            JSON data
        proxies: dict
            proxies to use in request (UNUSED)
        verbose: bool
            verbose output (default False)

        Returns
        -------
        response_headers: list
            response headers
        content: dict
            response content
        """
        # If no custom headers are given, use our own
        if headers is None:
            headers = {
                'Authorization': ' '.join([self.token_type, self.token]),
                'Accept-Encoding': 'gzip',
                'User-Agent': common.user_agent
            }

        # If uri_keys are given, find the uri from the uris dictionary
        if uri_keys is not None:
            uri = self._uri_root + common.uris[uri_keys[0]][uri_keys[1]]

        # Provide the input to the request
        if uri_args is not None:
            # uri	= uri % uri_args
            # Unencoded url does not work all of sudden...

            # encode_args = lambda x: urllib.parse.quote(x) if type(x) in (str, unicode) else x
            # if type(uri_args) in (list, tuple):
            # 	uri = uri % tuple(map(encode_args, uri_args))
            # else:
            # 	uri = uri % encode_args(uri_args)

            if type(uri_args) not in (list, tuple):
                uri_args = [uri_args]

            uri_args = [x if isinstance(x, Number) else urllib.parse.quote(x) for x in uri_args]
            uri = uri % tuple(uri_args)

        # Append query to the uri if body is given
        if body is not None:
            uri = uri + "?" + urllib.parse.urlencode(body)

        # Encode the data
        if data is not None:
            data = urllib.parse.urlencode(data).encode('utf-8')

        if jsondata is not None:
            headers['Content-Type'] = 'application/json'
            data = json.dumps(jsondata).encode('utf-8')

        # uri = uri.encode('utf-8')
        request = urllib.request.Request(uri, data=data, headers=headers)
        try:
            response = self.__send_request(request)
            status_code = response.getcode()
            if status_code != 200:
                print("HTTP status code: %d" % status_code)
                verbose = True
        except:
            (exc_type, exc_value, exc_trace) = sys.exc_info()
            if exc_type is ssl.CertificateError:
                print("A certificate verification error occurred for the request to '%s'. "
                      "Certificate verification is required by default, but can be disabled by "
                      "using the ignore_ssl_errors argument for the Connection constructor." % uri)
                raise
            if exc_type is urllib.error.HTTPError:
                message_bytes = exc_value.read()
                message_str = message_bytes.decode('utf-8')
                print(exc_value.msg + '\n Details: ' + message_str)
                raise

            print("Trying to reconnect the EMS API.")
            self.reconnect()
            print("Done.")
            response = self.__send_request(request)
        response_headers = response.getheaders()

        # If the response is compressed, decompress it.
        if response.info().get('Content-Encoding') == 'gzip':
            buffer = io.BytesIO(response.read())
            file = gzip.GzipFile(fileobj=buffer)
            content = json.loads(file.read())
        else:
            content = json.loads(response.read())

        if verbose:
            print("URL: %s" % response.geturl())
            pp.pprint(response_headers)
            pp.pprint(content)

        return response_headers, content

    def __send_request(self, req):
        """
        Sends the request and returns the response, optionally ignoring ssl errors.

        Parameters
        ----------
        req: urllib.request.Request
            request object

        Returns
        -------
        request: urllib.request.Request
            request object
        """
        # Normally you do NOT want to ignore SSL errors, but this is
        # sometimes necessary on beta API endpoints without a proper cert.
        # TODO: Find a real fix for GE Mac OS machines that are having trouble verifying.
        if self.__ignore_ssl_errors:
            try:
                request = urllib.request.urlopen(req, context=None)
            except urllib.error.URLError:
                request = urllib.request.urlopen(req, context=ssl._create_unverified_context())
        else:
            request = urllib.request.urlopen(req, context=None)
        return request


def _print_resp(resp):
    for r in resp:
        pp.pprint(r)



import pandas as pd
from .query import Query


class Profile(Query):
    """
    A class for performing profile queries in the EMS API.
    """
    def __init__(self, conn, ems_name, profile_number=None, profile_name='', searchtype='match'):
        """
        Parameters
        ----------
        conn: ems.connection.Connection
            A valid instantiated EMS connection object.
        ems_name: str
            A valid EMS system name.
        profile_number: int
            A profile number to select.
        profile_name: str
            A profile name.
        searchtype: str
            search type to perform:
                contain: uses the Pandas string method 'contains' on the column as string
                match: uses the Pandas string method 'match' on the column as string

        Raises
        ------
        ValueError
            If neither `profile_number` and `profile_name` are specified.
        LookupError
            If a profile matching either the `profile_number` or `profile_name` is not found.
        """
        Query.__init__(self, conn, ems_name)

        # Set parameters from arguments
        self._conn = conn
        self._ems_name = ems_name
        self._input_profile_number = profile_number
        self._input_profile_name = profile_name

        # To be filled out later
        self._guid = None
        self._glossary = None
        self._events_glossary = None
        self._exact_search = False

        # Get EMS id.
        self._ems_id = self.get_ems_id()

        # Set up template strings.
        self._no_matches_template = "Did not find any profiles that could match {0}: {1}"
        self._many_matches_template = "Found multiple profiles that could match {0}: {1}.\n\n{2}"

        # Check that at least one of the profile number/name arguments was passed.
        if profile_number is None and profile_name == '':
            raise ValueError("Either profile_number, profile_name, or both must be specified.")

        # Decide how to search
        self.searchtype = searchtype
        self._search_type = 'number' if profile_number is not None else 'name'
        # Perform a search using the specified options
        self._search()

    def get_glossary(self):
        """
        Method to get the glossary for the profile

        Returns
        -------
        pd.DataFrame
            A Pandas DataFrame containing a profile glossary.
        """
        if self._guid is not None:
            resp_h, dict_data = self._conn.request(
                uri_keys=('profile', 'glossary'),
                uri_args=(self._ems_id, self._guid)
            )
            a = pd.DataFrame.from_dict(dict_data)
            # convert the dictionaries values within the glossaryItems column into a new DataFrame
            b = a['glossaryItems'].apply(pd.Series)
            # concatenate a and b, drop the old glossaryItems column name
            c = pd.concat([a, b], axis=1).drop('glossaryItems', axis=1)
            self._glossary = c
            return self._glossary
        else:
            print("The search results did not return a profile matching the input profile name and "
                  "number on the given system. Please try to instantiate the profile object with "
                  "different arguments.")
            return

    def get_events_glossary(self):
        """
        Method to get the events glossary for the profile

        Returns
        -------
        pd.DataFrame
            A Pandas datafarme containing the events glossary
        """
        if self._guid is not None:
            _, dict_data = self._conn.request(
                uri_keys=('profile', 'events'),
                uri_args=(self._ems_id, self._guid)
            )
            data = pd.DataFrame(dict_data)
            data.set_index('id', inplace=True)
            self._events_glossary = data
            return self._events_glossary
        else:
            print("The search results did not return a profile matching the input profile name and "
                  "number on the given system. Please try to instantiate the profile object with "
                  "different arguments.")
            return

    def __query_profile_results(self, flight_id):
        if self._guid is not None:
            _, dict_data = self._conn.request(
                uri_keys=('profile', 'profile_results'),
                uri_args=(self._ems_id, flight_id, self._guid)
            )
            return dict_data

    def get_measurements(self, flight_id):
        """
        Get default event measurements for selected flight

        Parameters
        ----------
        flight_id: int
            flight record identifier

        Returns
        -------
        measurements: pd.DataFrame
            event default measurements for selected flight
        """
        profile_results = self.__query_profile_results(flight_id)
        measurements = pd.DataFrame(profile_results['measurements']).sort_values('itemId')
        glossary = self.__filter_glossary('measurement', 'default')
        measurements['name'] = measurements['itemId'].map(glossary['name'].to_dict())
        return measurements

    def get_timepoints(self, flight_id):
        """
        Get default event timepoints for selected flight

        Parameters
        ----------
        flight_id: int
            flight record identifier

        Returns
        -------
        measurements: pd.DataFrame
            event default timepoints for selected flight
        """
        profile_results = self.__query_profile_results(flight_id)
        timepoints = pd.DataFrame(profile_results['timepoints']).sort_values('itemId')
        glossary = self.__filter_glossary('timepoint', 'default')
        timepoints['name'] = timepoints['itemId'].map(glossary['name'].to_dict())
        return timepoints

    def get_events(self, flight_id):
        """
        Get event information for selected flight

        Parameters
        ----------
        flight_id: int
            flight record identifier

        Returns
        -------
        event_data: pd.DataFrame
            selected event field for selected flight
        """
        profile_results = self.__query_profile_results(flight_id)
        events = pd.DataFrame(profile_results['events'])

        # Grab event names and ID's from the glossary.
        event_name_details = self.__filter_glossary('event', 'eventSpecific')[['eventTypeId', 'name']]\
            .astype({'eventTypeId': int})\
            .rename(columns={'name': 'eventName'})
        # Merge in the event names so the user has something human-readable to work with.
        event_data = events\
            .merge(event_name_details, how='left', left_on='eventType', right_on='eventTypeId')\
            .drop(columns='eventTypeId')

        return event_data

    def __filter_glossary(self, record_type, scope):
        # Retrieve the profile glossary if it has not already been loaded.
        if self._glossary is None:
            self.get_glossary()
        glossary = self._glossary.loc[(self._glossary['recordType'] == record_type), :]
        glossary = glossary.loc[(self._glossary['scope'] == scope), :]
        glossary = glossary.set_index('itemId')
        return glossary

    def __set_profile_attributes(self, profile_data):
        if isinstance(profile_data, pd.DataFrame):
            profile_data = profile_data.iloc[0]
        self._guid = profile_data['id']
        self._current_version = profile_data['currentVersion']
        self._library = profile_data['library']
        self._profile_name = profile_data['name']
        self._local_id = profile_data['localId']

    def __request_results(self):
        # get values via API.
        _, dict_data = self._conn.request(
            uri_keys=('profile', 'search'),
            uri_args=self._ems_id,
            body={'search': self._input_profile_name}
        )
        data = pd.DataFrame.from_dict(dict_data)  # create a DataFrame from the response.
        self._search_results = data
        return data

    def __validate_search_results(self, res):
        if len(res) == 0:
            raise LookupError(self._no_matches_template.format(
                'profile name',
                self._input_profile_name
            ))

    def __filter_results(self, res):
        filtered = res  # default assignment
        # search can be of three different types: {name+exact, name+inexact, number}
        if self._search_type == 'name':
            if self.searchtype == 'contain':
                filtered = res.loc[res['name'].str.lower()
                                              .str.contains(self._input_profile_name.lower()), :]
            elif self.searchtype == 'match':
                filtered = res.loc[res['name'].str.lower()
                                              .str.match(self._input_profile_name.lower()), :]
        elif self._search_type == 'number':  # number search
            # locate rows where the localId matches the input profile number.
            filtered = res.loc[res['localId'] == self._input_profile_number]
        if isinstance(filtered, pd.Series):
            filtered = filtered.to_frame(filtered.name).T
        return filtered

    def __validate_filtered(self, filtered):
        # Check the filtered values.
        # If only one row is found, set the class attributes using this row.
        if filtered.shape[0] == 1:
            print('Found a profile with the supplied profile number and name.')
            print('Profile name: {0}, profile number: {1}.'.format(filtered['name'],
                                                                   filtered['localId']))
            self.__set_profile_attributes(filtered)
        # if more than one rows are found (shouldn't be possible) return the profile with
        # the shortest name
        else:
            # if no rows are found, return an error.
            if filtered.empty:
                # choose the right template for the returned error.
                if self._search_type == 'number':
                    raise LookupError(self._no_matches_template.format(
                        'profile number',
                        self._input_profile_number
                    ))
                elif self._search_type == 'name':
                    raise LookupError(self._no_matches_template.format(
                        'profile name',
                        self._input_profile_name
                    ))
            if self._search_type == 'name':
                # Get shortest name
                print('Found {0} profiles matching the supplied profile name.'
                      .format(filtered.shape[0]))
                print('Returning profile with shortest name.')
                filtered =\
                    filtered.loc[filtered['name'].str.len() == filtered['name'].str.len().min(), :]
                if filtered.shape[0] > 1:
                    raise LookupError(self._many_matches_template.format(
                        'profile name',
                        self._input_profile_name,
                        str(filtered.loc[:, ['localId', 'name']])
                    ))
                else:
                    print('Profile name: {0}, profile number: {1}.'.format(filtered['name'],
                                                                   filtered['localId']))
                    self.__set_profile_attributes(filtered)
            elif self._search_type == 'number':
                raise LookupError(self._many_matches_template.format(
                    'profile number',
                    self._input_profile_number,
                    str(filtered.loc[:, ['localId', 'name']])
                ))

    def _search(self):
        res = self.__request_results()
        self.__validate_search_results(res)
        filtered = self.__filter_results(res)
        self.__validate_filtered(filtered)



from __future__ import absolute_import
from .asset import Asset


class EMS(Asset):
    """
    Retrieves EMS system-specific info
    """

    def __init__(self, conn):
        """
        EMS object initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        """
        Asset.__init__(self, conn, "EMS")

    def update_list(self):
        """
        Method to update available EMS systems list

        Returns
        -------
        None
        """
        Asset.update_list(self, uri_keys=('ems_sys', 'list'))

    def ensure_loaded(self):
        """
        Method to ensure list is loaded in the object

        Returns
        -------
        None
        """
        if not (Asset.list_all(self)):
            self.update_list()

    def get_id(self, name=None):
        """
        A method to get an ems_id using a supplied ems_name.

        If name is None the code retrieves the list of available ems systems, if there is only 1
        available system the id for that system is returned, otherwise a list of available systems
        is returned.

        Parameters
        ----------
        name: str
            An EMS system name to get the ID for.

        Raises
        ------
            LookupError
                If no name is provided and the user has access to more than one ems system
            ValueError
                If no matching systems are found with the specified name
        """

        # Support using integer IDs directly
        if isinstance(name, int):
            return name

        self.ensure_loaded()
        if name is not None:
            ems_systems = self.search('name', name.upper(), searchtype="match")
            if ems_systems.empty:
                sys_names = self.list_all()['name'].to_list()
                raise ValueError(
                    'No matching systems found. You have access to: {0}'.format(sys_names))
            id = ems_systems.iloc[0]['id']
        else:
            ems_systems = self.list_all()
            if ems_systems.shape[0] == 1:
                id = ems_systems.iloc[0]['id']
            else:
                raise LookupError(
                    'Multiple ems systems found. Please select one from the available:\n{0}'
                    .format(ems_systems.loc[:, ['id', 'name']])
                )
        return id

    def get_name(self, id_val=None):
        """
        Method to get an EMS instance name from its id

        Parameters
        ----------
        id_val: int
            value id

        Returns
        -------
        name: str
            matching name
        """
        self.ensure_loaded()
        name = self.search('id', id_val, searchtype="match")['name'].tolist()
        return name if len(name) > 1 else name[0]



from __future__ import absolute_import
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import sys
import pickle
import warnings
import pandas as pd
import numpy as np

from emspy.query import *
from .query import Query


class TSeriesQuery(Query):
    """
    Query object for timeseries data
    """
    def __init__(self, conn, ems_name=None, data_file=LocalData.default_data_file):
        """
        Query initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        ems_name: str
            EMS system name
        data_file: str
            path to database file
        """
        Query.__init__(self, conn, ems_name)
        self._init_assets(data_file)
        self.reset()

    def _init_assets(self, data_file):
        # Query._init_assets(self)
        self.__analytic = Analytic(self._conn, self._ems_id, data_file)

    def reset(self):
        """
        Method to reset the queryset

        Returns
        -------
        None
        """
        self.__columns = list()
        self.__queryset = {'select': []}

    def select_ids(self, analytic_ids, names=None, descriptions=None, units=None, lookup=False):
        """
        A method for selecting parameters to query for by analytic_id.
        This method adds analytics and relevant information as provided.
        It can also query the EMS API for relevant information if requested.

        Parameters
        ----------
            analytic_ids: list of str
                a list of analytic_ids to select and query for
            names: list of str
                a list of names for each analytic_id in analytic_ids
            descriptions: list of str (optional)
                a list of descriptions for each analytic_id in analytic_ids
            units: list of str (optional)
                a list of units for each analytic_id in analytic_ids
            names: list of str (optional)
                a list of names for each analytic_id in analytic_ids
            lookup: bool (optional)
                whether or not to look up names, descriptions and units using the EMS API

        Returns
        -------
        None

        Raises
        ------
            ValueError
                If `analytic_ids`, `names`, `descriptions` (if specified), and `units`
                (if specified) are not of the same length.
            ValueError
                If `analytic_ids`, and `names` are not both defined, unless `lookup=True` is
                specified, in which case names can be None.
            TypeError
                If `analytic_ids` is a string, but `names`, `descriptions`, and `units` are
                specified as something other than strings.
        """

        # We require names if we are not going to look up metadata about the analytic_ids.
        # We could just allow adding analytic_ids, but the user will likely have to pair the
        # results up with names information anyway, so it might as well be provided to the
        # select_ids method for a more friendly experience.
        if names is None and not lookup:
            raise ValueError("If `names` is not specified, you must set `lookup`=True")

        # If the input parameters are both strings, then we will cast them into lists so we can
        # proceed as normal (i.e. loop through them [once])
        if isinstance(analytic_ids, str):
            analytic_ids = [analytic_ids]
            # if names is not None, and names is not a string type object, raise an error.
            if names is not None:
                if isinstance(names, str):
                    names = [names]
                else:
                    raise TypeError("If a string is passed for `analytic_ids`, "
                                    "`names` must be a string as well.")

            # if descriptions is not None, and descriptions is not a string type object,
            # raise an error.
            if descriptions is not None:
                if isinstance(descriptions, str):
                    descriptions = [descriptions]
                else:
                    raise ValueError("If a string is passed for `analytic_ids`, `descriptions`"
                                     " must be a string as well.")

            # if units is not None, and units is not a string type object, raise an error.
            if units is not None:
                if isinstance(units, str):
                    units = [units]
                else:
                    raise ValueError("If a string is passed for `analytic_ids`, `units` must be a "
                                     "string as well.")

        # calculate the length of each input list
        # descriptions and units aren't necessarily input lists (could be None), but they are
        # necessarily the same length as analytic_ids
        lengths = {'analytic_ids': len(analytic_ids)}
        if names is not None:
            lengths['names'] = len(names)
        if descriptions is not None:
            lengths['descriptions'] = len(descriptions)
        if units is not None:
            lengths['units'] = len(units)

        length_set = set(lengths.values())  # This is a set of unique lengths

        # If there is more than one unique iterable length, then one must be wrong. Return an error.
        if len(length_set) > 1:
            error_str = ''
            for list_name, list_length in lengths.items():
                error_str = error_str + '\t\t{0:15}: {1}\n'.format(list_name, str(list_length))
            raise ValueError("All lists must be of the same length. Found lengths:\n{0}"
                             .format(error_str))

        # Fill out lists with empty strings that are the same length as the analytic_ids list,
        # if parameters are None.
        names = [''] * len(analytic_ids) if names is None else names
        descriptions = [''] * len(analytic_ids) if descriptions is None else descriptions
        units = [''] * len(analytic_ids) if units is None else units

        # loop over items
        for analytic_id, name, description, unit in zip(analytic_ids, names, descriptions, units):
            # if we aren't performing a lookup, just stuff in the lists that have been passed in.
            if not lookup:
                prm = {
                    'id': analytic_id,
                    'name': name,
                    'description': description,
                    'units': unit,
                    'ems_id': self._ems_id
                }
            # if we are performing a lookup, we need to make an API call to get relevant data.
            else:
                prm = self.__analytic.get_param_details(analytic_id)
                prm['ems_id'] = self._ems_id
            df = pd.DataFrame(prm, index=[0])

            self.__analytic._param_table = \
                self.__analytic._param_table.append(df, ignore_index=True, sort=True)

            # Put the param into JSON query string
            self.__queryset['select'].append({'analyticId': prm['id']})
            # Just in case you want to check what params are selected
            self.__columns.append(prm)
            self.__analytic._save_paramtable()

    def select(self, *args, **kwargs):
        """
        A method for selecting parameters to query for by name. This method searches for parameters
        by name and adds them to the current query.
        If multiple matches for a given parameter are found, the match with the shortest name
        is selected.

        Parameters
        ----------
            *args: `list` of `str`
                Variable length argument list of parameter names.
            *kwargs
                force_search: bool. If True, it searches for the parameter every time and ignores the local cache
        """

        keywords = args
        save_table = False

        for kw in keywords:
            if not 'force_search' in kwargs:
                # Get the param from param table
                prm = self.__analytic.get_param(kw)
            else:
                if kwargs.get("force_search"):
                    # Make an empty parameter to force the search
                    prm = dict(ems_id="", id="", name="", description="", units="")
                else:
                    # Get the param from param table
                    prm = self.__analytic.get_param(kw)
            if prm['id'] == "":
                # If the param's not found, call EMS API
                res_df = self.__analytic.search_param(kw, in_df=True)
                res_df['ems_id'] = self._ems_id
                # The first one is with the shortest name string. Pick that.
                prm = res_df.iloc[0, :].to_dict()
                # Add the new parameters to the param table for later uses
                self.__analytic._param_table = \
                    self.__analytic._param_table.append(res_df, ignore_index=True, sort=True)
                save_table = True

            # Put the param into JSON query string
            self.__queryset['select'].append({'analyticId': prm['id']})
            # Just in case you want to check what params are selected
            self.__columns.append(prm)
        if save_table:
            self.__analytic._save_paramtable()

    def range(self, start=None, end=None):
        """
        Sets the range for the timeseries

        Parameters
        ----------
        start: str
            start datetime
        end: str
            end datetime

        Returns
        -------
        None
        """
        self.__queryset['start'] = start
        self.__queryset['end'] = end

    def timepoint(self, tpoint):
        """
        Adds a time point offset to the queryset

        Parameters
        ----------
        tpoint: str
            timepoint datetime

        Returns
        -------
        None
        """
        if type(tpoint) == np.ndarray:
            tpoint = tpoint.tolist()
        self.__queryset['offsets'] = tpoint

    def run(self, flight, start=None, end=None, timestep=None, timepoint=None,
            discretes_as_strings=True):
        """
        Runs the query

        Parameters
        ----------
        flight: int
            flight record id
        start: str
            start datetime
        end: str
            end datetime
        timestep: int
            time step magnitude
        timepoint: str
            time point datetime
        discretes_as_strings: bool
            allows user to treat discrete values as strings (default True)

        Returns
        -------
        pd.DataFrame
            pandas dataframe with the selected data
        """

        # if start is None:
        #     start = 0.0

        # if end is None:
        #     end = self.flight_duration(flight)

        # if timestep is not None:
        #     timepoint = np.arange(start, end, timestep)

        # if timepoint is not None:
        #     self.timepoint(timepoint)
        # else:
        #     self.range(start, end)
        if timepoint is not None:
            self.timepoint(timepoint)

        # Allow the user to specify discretesAsStrings = 'false' per the EMS API documentation,
        # discretesAsStrings is true by default.
        # So it will only get set to false if that option is specified.
        # This will allow users to return discretes in time series parameters as int values.
        if not discretes_as_strings:
            self.__queryset['discretesAsStrings'] = 'false'
        elif timestep is not None:
            start = 0.0 if start is None else start
            if end is None:
                raise ValueError("End timepoint should be given along with timestep input.")
            self.timepoint(np.arange(start, end+1e-10, timestep))
        else:
            self.range(start, end)

        _, content = self._conn.request(
            uri_keys=("analytic", "query"),
            uri_args=(self._ems_id, flight),
            jsondata=self.__queryset
        )

        if 'message' in content:
            sys.exit('API query for flight %d was unsuccessful.\n'
                     'Here is the message from API: %s' % (flight, content['message']))
        
        # Put the data in Pandas DataFrame
        df = pd.DataFrame({"Time (sec)": content['offsets']})

        for i, prm in enumerate(self.__columns):
            df[prm['name']] = content['results'][i]['values']

        # "\r\x1b[K" is overwrite print.
        return df

    def multi_run(self, flight, start=None, end=None, timestep=None, timepoint=None, save_file=None,
                  verbose=True):
        """
        Wraps the run method to run several timepoints

        Parameters
        ----------
        flight: int
            flight record id
        start: str
            start datetime
        end: str
            end datetime
        timestep: int
            time step magnitude
        timepoint: str
            time point datetime
        save_file: str
            path to dump the data to using pickle
        verbose: bool
            sets verbose output

        Returns
        -------
        list
            list of dataframes for each run
        """
        res = list()
        attr_flag = False
        if isinstance(flight, pd.DataFrame):
            flight_record = flight["Flight Record"]
            attr_flag = True
        else:
            flight_record = flight

        # param processing
        if start is None:
            start = [None] * len(flight_record)
        if end is None:
            end = [None] * len(flight_record)
        if not hasattr(timestep, "__len__"):
            timestep = [timestep] * len(flight_record)
        if timepoint is None:   
            timepoint = [None] * len(flight_record)
        else: 
            warnings.warn("Time points are not yet supported. "
                          "The given time points will be ignored.")
            timepoint = [None] * len(flight_record)
            
        if verbose:
            print('\n=== Start running time-series data querying for %d flights ===\n'
                  % len(flight_record))
        
        for i, fr in enumerate(flight_record):
            if verbose:
                print('\r\x1b[K%d / %d: FR %d' % (i + 1, len(flight_record), fr), end=' ')
            i_res = dict()
            if attr_flag:
                i_res['flt_data'] = flight.iloc[i, :].to_dict()
            else:
                i_res['flt_data'] = {'Flight Record': fr}
            i_res['ts_data'] = self.run(fr, start[i], end[i], timestep[i])
            res.append(i_res)
            if save_file is not None:
                pickle.dump(res, open(save_file, 'wb'))
        if verbose:
            print('Done')
        return res

    def flight_duration(self, flight, unit="second"):
        """
        Deprecated

        Parameters
        ----------
        flight: str
            flight id
        unit: str
            units

        Returns
        -------
        float
            flight duration
        """
        p = self.__analytic.get_param("hours of data (hours)")
        if p["id"] == "":
            res_df = self.__analytic.search_param("hours of data (hours)", in_df=True)
            p = res_df.iloc[0].to_dict()
            self.__analytic._param_table = \
                self.__analytic._param_table.append(res_df, ignore_index=True, sort=True)
            self.__analytic._save_paramtable()
        q = {
            "select": [{"analyticId": p["id"]}],
            "size": 1
        }

        _, content = self._conn.request(
            uri_keys=("analytic", "query"),
            uri_args=(self._ems_id, flight),
            jsondata=q
        )
        if 'message' in content:
            sys.exit('API query for flight %d, parameter = "%s" was unsuccessful.\n'
                     'Here is the message from API: %s' % (flight, p['name'], content['message']))
        fl_len = content['results'][0]['values'][0]

        if unit == "second":
            t = fl_len * 60 * 60
        elif unit == "minute":
            t = fl_len * 60
        elif unit == "hour":
            t = fl_len
        else:
            sys.exit("Unrecognizable time unit (%s)." % unit)
        return t

    def get_queryset(self):
        """
        Returns a dictionary with the queryset for the current time series query.

        Returns:
            list: a list describing the current queryset.
        """
        return self.__queryset



from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import re
import urllib.error
from .asset import Asset
import pandas as pd


class AnalyticSet(Asset):
    """
    Manages analytic sets querying
    """

    def __init__(self, conn, ems_id, exclude_trees=[], exclude_patterns=[]):
        """
        Analytics set asset object initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        ems_id: int
            EMS system id
        exclude_trees: list
            list of trees to exclude from the group update, such as:
            "Engine Services", "ADI", "EMS Library"
        exclude_patterns: list
            list of regex patterns for exclusion of groups, such as:
            r"Engine Services:\d{9}.*" to exclude all SSO subdirectories in Engine Services
        """
        Asset.__init__(self, conn, "AnalyticSet")
        self._ems_id = ems_id
        self._analytic_sets = []
        self._analytics = []
        self.exclude_trees = exclude_trees
        self.exclude_patterns = exclude_patterns
        if len(self.exclude_patterns) > 0:
            self.exclude_patterns = [re.compile(pattern) for pattern in self.exclude_patterns]
        self.update_list()

    def __get_analytic_set_group(self, groupId):
        if groupId in self.exclude_trees:
            print('-- Excluding analytic set group {0}'.format(groupId))
            return
        for pattern in self.exclude_patterns:
            if re.match(pattern, groupId) is not None:
                print('-- Excluding analytic set group {0}'.format(groupId))
                return
        try:
            print('-- Fetching analytic set group {0}'.format(groupId))
            _, dict_data = self._conn.request(
                uri_keys=('analyticSet', 'analytic_set_group'),
                uri_args=(self._ems_id, groupId)
            )
            self._analytic_sets.append(dict_data)
            for group in dict_data['groups']:
                self.__get_analytic_set_group(group['groupId'])
        except urllib.error.HTTPError:
            print('-- Failed to fetch analytic set group {0}'.format(groupId))

    def update_list(self):
        """
        Update analytic sets list

        Returns
        -------
        None
        """
        self.__get_analytic_set_group('Root')
        self._analytic_sets = pd.DataFrame(self._analytic_sets)
        self._analytic_sets = self._analytic_sets.drop(['groups', 'sets'], axis=1)\
            .join(self._analytic_sets['sets'].apply(pd.Series))\
            .melt(id_vars=['name', 'groupId'], value_name='set')\
            .drop('variable', axis=1)\
            .dropna(subset=['set'])
        self._analytic_sets['set'] = self._analytic_sets['set'].apply(lambda d: d['name'])
        self._analytic_sets.sort_values(['name', 'groupId', 'set'], inplace=True)
        self._analytic_sets.reset_index(drop=True, inplace=True)

    def select_set(self, analytics_set_name=None, searchtype='match'):
        """
        Get fleet name from id

        Parameters
        ----------
        analytics_set_name: str
            set name
        searchtype: str
            search type:
                contains: inexact matching (returns shortest)
                match: exact matching

        Returns
        -------
        None
        """
        analytics_set = self.search('set', analytics_set_name, searchtype)
        if analytics_set.shape[0] > 1:
            analytics_set = analytics_set.loc[analytics_set['set'].str.len().idxmin(), :]\
                .to_frame().T
        _, dict_data = self._conn.request(
            uri_keys=('analyticSet', 'analytic_set'),
            uri_args=(self._ems_id, analytics_set.iloc[0]['groupId'], analytics_set.iloc[0]['set'])
        )
        self._analytics = pd.DataFrame(dict_data)
        self._analytics = self._analytics.drop('items', axis=1) \
            .join(self._analytics['items'].apply(pd.Series))
        self._analytics = self._analytics.drop('analytic', axis=1) \
            .join(self._analytics['analytic'].apply(pd.Series), rsuffix='_analytic')
        return self._analytics

    def list_all(self):
        return self._analytics

    def _rename_datacol(self, old, new):
        # Rename a column
        raise NotImplementedError

    def data_colnames(self):
        """
        Returns the columns for the asset dataframe

        Returns
        -------
        pd.Index
            dataframe columns
        """
        return self._analytics.columns




import os
import sqlite3
import sys
from builtins import object

import numpy as np
import pandas as pd

import emspy


class LocalData(object):
	"""
	Class to save metadata in a local DB file
	"""
	default_data_file = os.path.join(emspy.__path__[0], "data", "emsMetaData.db")
	table_info = {
		"fieldtree": ["uri_root", "ems_id", "db_id", "id", "nodetype", "type", "name", "parent_id"],
		"dbtree": ["uri_root", "ems_id", "id", "nodetype", "name", "parent_id"],
		"kvmaps": ["uri_root", "ems_id", "id", "key", "value"],
		"params": ["uri_root", "ems_id", "id", "name", "description", "units"]
	}

	def __init__(self, dbfile=default_data_file):
		"""
		LocalData object initialization

		Parameters
		----------
		dbfile: str
			path to database file
		"""
		self.__dbfile = None
		self._conn = None

		# This is done to support a mode that does not use a data file at all
		# (when dbfile is set to None explicitly)
		if dbfile:
			self.__dbfile = os.path.abspath(dbfile)
			self.__connect()

	def __del__(self):
		self.close()

	def __connect(self):
		if self.__dbfile:
			self._conn = sqlite3.connect(self.__dbfile)

	@staticmethod
	def __check_colnames(table_name, df):
		colnames = np.array(LocalData.table_info[table_name])
		chk_cols = np.array([c in df.columns for c in colnames])
		missing = colnames[~chk_cols]
		if any(~chk_cols):
			sys.exit("Data misses the following columns that are required: %s" % missing)

	def close(self):
		"""
		Closes SQLite connection

		Returns
		-------
		None
		"""
		if self._conn is not None:
			self._conn.close()

	def append_data(self, table_name, df):
		"""
		Adds data to a table

		Parameters
		----------
		table_name: str
			table to append to
		df: pandas.DataFrame
			data to append

		Returns
		-------
		None
		"""
		self.__check_colnames(table_name, df)
		if self.__dbfile is not None:
			df.to_sql(table_name, self._conn, index=False, if_exists="append")

	def get_data(self, table_name, condition=None):
		"""
		Selects data from a table

		Parameters
		----------
		table_name: str
			table to query
		condition: str
			filtering condition (WHERE clause)

		Returns
		-------
		pd.DataFrame
			selected data from the table
		"""
		if (self.__dbfile is not None) and self.table_exists(table_name):
			query = "SELECT * FROM %s" % table_name
			if condition is not None:
				query = query + " WHERE %s" % condition
			query = query + ";"
			df = pd.read_sql_query(query, self._conn)

			# Strange columns appear. Get only the actual columns
			return df[[col for col in LocalData.table_info[table_name] if col in df]]
		return pd.DataFrame(columns=LocalData.table_info[table_name])

	def delete_data(self, table_name, condition=None):
		"""
		Drops a table or deletes from a table

		Parameters
		----------
		table_name: str
			table to drop or to remove records from
		condition: str
			removal condition (WHERE clause), if set to None (default) the table is dropped instead

		Returns
		-------
		None
		"""
		if (self.__dbfile is not None) and self.table_exists(table_name):
			if condition is None:
				self._conn.execute("DROP TABLE %s" % table_name)
			else:
				self._conn.execute("DELETE FROM %s WHERE %s;" % (table_name, condition))
			self._conn.commit()

	def delete_all_tables(self):
		"""
		Deletes all tables from the database

		Returns
		-------
		None
		"""
		if self.__dbfile is not None:
			for table_name in list(LocalData.table_info.keys()):
				if self.table_exists(table_name):
					self._conn.execute("DROP TABLE %s" % table_name)
			self._conn.commit()

	def table_exists(self, table_name):
		"""
		Checks if a table exists

		Parameters
		----------
		table_name: str
			table name

		Returns
		-------
		bool
			False if the table does not exist, True otherwise.
			Also returns False if there is no database file
		"""
		if self.__dbfile is None:
			return False

		cursor = self._conn.cursor()
		cursor.execute("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name")
		tables = [t[0] for t in cursor.fetchall()]
		return table_name in tables

	def file_loc(self):
		"""
		Get database path

		Returns
		-------
		str
			database file path
		"""
		return self.__dbfile

	def is_db_path_correct(self, path):
		"""
		Checks if the path represents the current data file.

		Parameters
		----------
		path: str
			The path to compare.

		Returns
		-------
		bool
			True if the path is the current data file, False otherwise.
		"""
		if path is None:
			return self.file_loc() is None
		else:
			return self.file_loc() == os.path.abspath(path)



uri_root = {
    'prod': 'https://ems.efoqa.com/api',
    'cluster': 'https://ceod.efoqa.com/api',
    'stable': 'https://emsapi.ausdig.com/api',
    'beta': 'https://emsapibeta.ausdig.com/api',
    'nightly': 'https://emsapitest.ausdig.com/api'
}

user_agent = 'ems-api-sdk Python v0.2'

uris = {
    'sys': {
        'auth': '/token'
    },
    'ems_sys': {
        'list': '/v2/ems-systems',  # (ems-system_id)
        'ping': '/v2/ems-systems/%s/ping',  # (ems-system_id)
        'info': '/v2/ems-systems/%s/info'
    },
    'fleet': {
        'list': '/v2/ems-systems/%s/assets/fleets',  # (ems-system_id)
        'info': '/v2/ems-systems/%s/assets/fleets/%s'  # (ems-system_id, fleet_id)
    },
    'aircraft': {
        'list': '/v2/ems-systems/%s/assets/aircraft',  # (ems-system_id)
        'info': '/v2/ems-systems/%s/assets/aircraft/%s'  # (ems-system_id, aircraft_id)
    },
    'flt_phase': {
        'list': '/v2/ems-systems/%s/assets/flight-phases',  # (ems-system_id)
        'info': '/v2/ems-systems/%s/assets/flight-phases/%s'  # (ems-system_id, flt_phase_id)
    },
    'airport': {
        'list': '/v2/ems-systems/%s/assets/airports',  # (ems-system_id)
        'info': '/v2/ems-systems/%s/assets/airports/%s'  # (ems-system_id, airport_id)
    },
    'database': {
        'group': '/v2/ems-systems/%s/database-groups',  # (ems-system_id)
        'field_group': '/v2/ems-systems/%s/databases/%s/field-groups',  # (ems-system_id, data_src_id)
        'field': '/v2/ems-systems/%s/databases/%s/fields/%s',  # (ems-system_id, database_id, field_id)
        'query': '/v2/ems-systems/%s/databases/%s/query',
        'create': '/v2/ems-systems/%s/databases/%s/create',  # (emsSystemId, databaseId)
        'open_asyncq': '/v2/ems-systems/%s/databases/%s/async-query',  # (ems-system_id, database_id)
        'get_asyncq': '/v2/ems-systems/%s/databases/%s/async-query/%s/read/%s/%s',  # (ems-system_id, database_id, async_query_id, start_row, end_row)
        'close_asyncq': '/v2/ems-systems/%s/databases/%s/async-query/%s'
    },
    'analytic': {
        'search': '/v2/ems-systems/%s/analytics',  # (emsSystemId)
        'search_f': '/v2/ems-systems/%s/flights/%s/analytics',  # (emsSystemId, flightId)
        'group': '/v2/ems-systems/%s/analytic-groups',  # (emsSystemId)
        'group_f': '/v2/ems-systems/%s/flights/%s/analytic-groups',  # (emsSystemId, flightId)
        'query': '/v2/ems-systems/%s/flights/%s/analytics/query',  # (emsSystemId, flightId)
        'metadata': '/v2/ems-systems/%s/flights/%s/analytics/metadata'  # (emsSystemId, flightId)
    },
    'profile': {
        'profile_results': '/v2/ems-systems/%s/flights/%s/profiles/%s/query',  # (emsSystemId, flightId, profileId)
        'search': '/v2/ems-systems/%s/profiles',  # (emsSystemId)
        'glossary': '/v2/ems-systems/%s/profiles/%s/glossary',  # (emsSystemId, profileId)
        'events': '/v2/ems-systems/%s/profiles/%s/events',  # (emsSystemId, profileId)
        'single_event': '/v2/ems-systems/%s/profiles/%s/events/%s',  # (emsSystemId, profileId, eventId)
    },
    'analyticSet': {
        'analytic_set_groups': '/v2/ems-systems/%s/analytic-set-groups',  # (emsSystemId),
        'analytic_set_group': '/v2/ems-systems/%s/analytic-set-groups/%s',  # (emsSystemId, groupId)
        'analytic_set': '/v2/ems-systems/%s/analytic-set-groups/%s/analytic-sets/%s',  # (emsSystemId, groupId, analyticSetName)
    }
}



from builtins import object
import pandas as pd


class Asset(object):
    """
    Base class of all "asset" classes and the EMS class. It is a template class so
    should not be instantiated directly.
    """

    def __init__(self, conn, asset_type):
        """
        Asset initialization

        Parameters
        ----------
        conn: emspy.connection.Connection
            connection object
        asset_type: str
            asset identifier
        """
        self._conn = conn
        self._asset_type = asset_type
        self._assets = None

    def update_list(self, uri_keys=None, uri_args=None, body=None, colsort=True):
        """
        Base method to update an asset list

        Parameters
        ----------
        uri_keys: tuple
            request uri keys
        uri_args: list, tuple
            request uri arguments
        body: dict
            request body
        colsort: bool
            Whether to sort the columns or not

        Returns
        -------
        None
        """
        _, dict_data = self._conn.request(uri_keys=uri_keys, uri_args=uri_args, body=body)
        assets = pd.DataFrame.from_dict(dict_data)
        if colsort:
            assets = assets[sorted(assets.columns, key=len)]
        self._assets = assets

    def list_all(self):
        """
        Method to list all assets

        Returns
        -------
        list
            listed assets
        """
        return self._assets

    def search(self, col, val=None, searchtype="contain"):
        """
        Method to search for a value in a column

        Parameters
        ----------
        col: str
            column to search in
        val: str
            value to search
        searchtype: str
            search type to perform:
                contain: uses the Pandas string method 'contains' on the column as string
                match: uses the Pandas string method 'match' on the column as string

        Returns
        -------
        str
            returns the found value
        """
        if val is None:
            return self.list_all()
        assets = self._assets
        if type(val) == str:
            if searchtype == "contain":
                return assets[assets[col].str.contains(val, na=False, case=False)]
            if searchtype == "match":
                return assets[assets[col].str.match(val, case=False)]
        return assets[assets[col] == val]

    def get_id(self, name=None, keyword=None):
        """
        Base method for getting an ID
        """
        raise NotImplementedError()

    def get_name(self, id_val=None):
        """
        Base method for getting a name from an ID
        """
        raise NotImplementedError()

    def _rename_datacol(self, old, new):
        # Rename a column
        self._assets = self._assets.rename(columns={old: new})

    def data_colnames(self):
        """
        Returns the columns for the asset dataframe

        Returns
        -------
        pd.Index
            dataframe columns
        """
        return self._assets.columns


def size(an_asset):
    if not issubclass(an_asset, Asset):
        raise TypeError("Input should be an object inherited from Asset class.")
    if an_asset.list_all() is None:
        return 0
    else:
        return an_asset.list_all().shape[0]
